from lenstronomy.SimulationAPI.data_api import DataAPI
import lenstronomy.Util.util as util
import numpy as np
import unittest


class TestDataAPI(object):
    def setup_method(self):
        numpix = 10
        self.ccd_gain = 4.0
        self.pixel_scale = 0.13
        self.read_noise = 10.0
        kwargs_instrument = {
            "read_noise": self.read_noise,
            "pixel_scale": self.pixel_scale,
            "ccd_gain": self.ccd_gain,
        }

        exposure_time = 100
        sky_brightness = 20.0
        self.magnitude_zero_point = 21.0
        num_exposures = 2
        seeing = 0.9
        kwargs_observations = {
            "exposure_time": exposure_time,
            "sky_brightness": sky_brightness,
            "magnitude_zero_point": self.magnitude_zero_point,
            "num_exposures": num_exposures,
            "seeing": seeing,
            "psf_type": "GAUSSIAN",
            "kernel_point_source": None,
        }
        self.kwargs_data = util.merge_dicts(kwargs_instrument, kwargs_observations)
        self.api = DataAPI(numpix=numpix, data_count_unit="ADU", **self.kwargs_data)

        kwargs_observations = {
            "exposure_time": exposure_time,
            "sky_brightness": sky_brightness,
            "magnitude_zero_point": self.magnitude_zero_point,
            "num_exposures": num_exposures,
            "seeing": seeing,
            "psf_type": "PIXEL",
            "kernel_point_source": np.ones((3, 3)),
        }
        kwargs_data = util.merge_dicts(kwargs_instrument, kwargs_observations)
        self.api_pixel = DataAPI(numpix=numpix, data_count_unit="ADU", **kwargs_data)

        self.ra_at_xy_0 = 0.02
        self.dec_at_xy_0 = 0.02
        self.transform_pix2angle = [[-self.pixel_scale, 0], [0, self.pixel_scale]]
        kwargs_pixel_grid = {
            "ra_at_xy_0": self.ra_at_xy_0,
            "dec_at_xy_0": self.dec_at_xy_0,
            "transform_pix2angle": self.transform_pix2angle,
        }
        self.api_pixel_grid = DataAPI(
            numpix=numpix,
            kwargs_pixel_grid=kwargs_pixel_grid,
            data_count_unit="ADU",
            **self.kwargs_data
        )

    def test_data_class(self):
        data_class = self.api.data_class
        assert data_class.pixel_width == self.pixel_scale

    def test_psf_class(self):
        psf_class = self.api.psf_class
        assert psf_class.psf_type == "GAUSSIAN"
        psf_class = self.api_pixel.psf_class
        assert psf_class.psf_type == "PIXEL"

    def test_kwargs_data(self):
        kwargs_data = self.api.kwargs_data
        assert kwargs_data["ra_at_xy_0"] != self.ra_at_xy_0
        kwargs_data = self.api_pixel_grid.kwargs_data
        assert kwargs_data["ra_at_xy_0"] == self.ra_at_xy_0


class TestRaise(unittest.TestCase):
    def test_raise(self):
        numpix = 10
        self.ccd_gain = 4.0
        self.pixel_scale = 0.13
        self.read_noise = 10.0
        kwargs_instrument = {
            "read_noise": self.read_noise,
            "pixel_scale": self.pixel_scale,
            "ccd_gain": self.ccd_gain,
        }

        exposure_time = 100
        sky_brightness = 20.0
        magnitude_zero_point = 21.0
        num_exposures = 2
        seeing = 0.9
        kwargs_observations = {
            "exposure_time": exposure_time,
            "sky_brightness": sky_brightness,
            "magnitude_zero_point": magnitude_zero_point,
            "num_exposures": num_exposures,
            "seeing": seeing,
            "psf_type": "wrong",
            "kernel_point_source": None,
        }
        kwargs_data = util.merge_dicts(kwargs_instrument, kwargs_observations)
        data_api = DataAPI(numpix=numpix, data_count_unit="ADU", **kwargs_data)
        print(data_api._psf_type)
        with self.assertRaises(ValueError):
            data_api = DataAPI(numpix=numpix, data_count_unit="ADU", **kwargs_data)
            psf_class = data_api.psf_class

        kwargs_observations = {
            "exposure_time": exposure_time,
            "sky_brightness": sky_brightness,
            "magnitude_zero_point": magnitude_zero_point,
            "num_exposures": num_exposures,
            "seeing": seeing,
            "psf_type": "PIXEL",
            "kernel_point_source": None,
        }
        kwargs_data = util.merge_dicts(kwargs_instrument, kwargs_observations)
        with self.assertRaises(ValueError):
            data_api = DataAPI(numpix=numpix, data_count_unit="ADU", **kwargs_data)
            psf_class = data_api.psf_class

        kwargs_data["kernel_point_source"] = np.ones((3, 3))
        kwargs_pixel_grid = {"ra_at_xy_0": 0.02, "dec_at_xy_0": 0.02}
        with self.assertRaises(ValueError):
            data_api = DataAPI(
                numpix=numpix, kwargs_pixel_grid=kwargs_pixel_grid, **kwargs_data
            )



"""Created on Sep 30, 2013 modified on March 3-7, 2020.

@authors: J. Akeret, S. Birrer, A. Shajib
"""

from copy import copy
from math import floor
import math
import numpy as np

__all__ = ["ParticleSwarmOptimizer"]


class ParticleSwarmOptimizer(object):
    """Optimizer using a swarm of particles.

    :param func:
        A function that takes a vector in the parameter space as input and
        returns the natural logarithm of the posterior probability for that
        position.

    :param low: array of the lower bound of the parameter space
    :param high: array of the upper bound of the parameter space
    :param particle_count: the number of particles to use.
    :param pool: (optional)
        An alternative method of using the parallelized algorithm. If
        provided, the value of ``threads`` is ignored and the
        object provided by ``pool`` is used for all parallelization. It
        can be any object with a ``map`` method that follows the same
        calling sequence as the built-in ``map`` function.
    """

    def __init__(
        self, func, low, high, particle_count=25, pool=None, args=None, kwargs=None
    ):
        """

        :param func: function to call to return log likelihood
        :type func: python definition
        :param low: lower bound of the parameters
        :type low: numpy array
        :param high: upper bound of the parameters
        :type high: numpy array
        :param particle_count: number of particles in each iteration of the PSO
        :type particle_count: int
        :param pool: MPI pool for mapping different processes
        :type pool: None or MPI pool
        :param args: positional arguments to send to `func`. The function
        will be called as `func(x, *args, **kwargs)`.
        :type args: `list`
        :param kwargs: keyword arguments to send to `func`. The function
        will be called as `func(x, *args, **kwargs)`
        :type kwargs: `dict`
        """
        self.low = [l for l in low]
        self.high = [h for h in high]
        self.particleCount = particle_count
        self.pool = pool

        self.param_count = len(self.low)

        self.swarm = self._init_swarm()
        self.global_best = Particle.create(self.param_count)

        self.func = _FunctionWrapper(func, args, kwargs)

    def __getstate__(self):
        """In order to be generally pickleable, we need to discard the pool object
        before trying."""
        d = self.__dict__
        d["pool"] = None
        return d

    def __setstate__(self, state):
        self.__dict__ = state

    def set_global_best(self, position, velocity, fitness):
        """Set the global best particle.

        :param position: position of the new global best
        :type position: `list` or `ndarray`
        :param velocity: velocity of the new global best
        :type velocity: `list` or `ndarray`
        :param fitness: fitness of the new global best
        :type fitness: `float`
        :return: `None`
        :rtype:
        """
        self.global_best.position = [p for p in position]
        self.global_best.velocity = [v for v in velocity]
        self.global_best.fitness = fitness

    def _init_swarm(self):
        """Initiate the swarm.

        :return:
        :rtype:
        """
        swarm = []
        for _ in range(self.particleCount):
            swarm.append(
                Particle(
                    np.random.uniform(self.low, self.high, size=self.param_count),
                    np.zeros(self.param_count),
                )
            )

        return swarm

    def sample(
        self,
        max_iter=1000,
        c1=1.193,
        c2=1.193,
        p=0.7,
        m=1e-3,
        n=1e-2,
        early_stop_tolerance=None,
        verbose=True,
    ):
        """Launches the PSO. Yields the complete swarm per iteration.

        :param max_iter: maximum iterations
        :param c1: cognitive weight
        :param c2: social weight
        :param p: stop criterion, percentage of particles to use
        :param m: stop criterion, difference between mean fitness and global best
        :param n: stop criterion, difference between norm of the particle vector and
            norm of the global best
        :param early_stop_tolerance: will terminate at the given value (should be
            specified as a chi^2)
        :param verbose: prints when it stopped
        :type verbose: boolean
        """

        self._get_fitness(self.swarm)
        i = 0
        while True:
            for particle in self.swarm:
                if self.global_best.fitness < particle.fitness:
                    self.global_best = particle.copy()
                    # if(self.isMaster()):
                    # print("new global best found %i %s"%(i,
                    # self.global_best.__str__()))

                if particle.fitness > particle.personal_best.fitness:
                    particle.update_personal_best()

            if i >= max_iter:
                if self.is_master():
                    if verbose:
                        print("Max iteration reached! Stopping.")
                return

            if self._converged(i, p=p, m=m, n=n):
                if self.is_master():
                    if verbose:
                        print("Converged after {} iterations!".format(i))
                        print(
                            "Best fit found: ",
                            self.global_best.fitness,
                            self.global_best.position,
                        )
                return

            if early_stop_tolerance is not None:
                if self._acceptable_convergence(early_stop_tolerance):
                    return

            for particle in self.swarm:
                w = 0.5 + np.random.uniform(0, 1, size=self.param_count) / 2
                # w=0.72
                part_vel = w * np.array(particle.velocity)
                cog_vel = (
                    c1
                    * np.random.uniform(0, 1, size=self.param_count)
                    * (
                        np.array(particle.personal_best.position)
                        - np.array(particle.position)
                    )
                )
                soc_vel = (
                    c2
                    * np.random.uniform(0, 1, size=self.param_count)
                    * (
                        np.array(self.global_best.position)
                        - np.array(particle.position)
                    )
                )
                particle.velocity = (part_vel + cog_vel + soc_vel).tolist()
                particle.position = (
                    np.array(particle.position) + np.array(particle.velocity)
                ).tolist()

            self._get_fitness(self.swarm)

            swarm = []
            for particle in self.swarm:
                swarm.append(particle.copy())
            yield swarm

            i += 1

    def optimize(
        self,
        max_iter=1000,
        verbose=True,
        c1=1.193,
        c2=1.193,
        p=0.7,
        m=1e-3,
        n=1e-2,
        early_stop_tolerance=None,
    ):
        """Run the optimization and return a full list of optimization outputs.

        :param max_iter: maximum iterations
        :param verbose: if `True`, print a message every 10 iterations
        :param c1: cognitive weight
        :param c2: social weight
        :param p: stop criterion, percentage of particles to use
        :param m: stop criterion, difference between mean fitness and global best
        :param n: stop criterion, difference between norm of the particle
         vector and norm of the global best
        :param early_stop_tolerance: will terminate at the given value (should be specified as a chi^2)
        """
        log_likelihood_list = []
        vel_list = []
        pos_list = []

        num_iter = 0
        for _ in self.sample(max_iter, c1, c2, p, m, n, early_stop_tolerance):
            log_likelihood_list.append(self.global_best.fitness)
            vel_list.append(self.global_best.velocity)
            pos_list.append(self.global_best.position)
            num_iter += 1

            if verbose and self.is_master():
                if num_iter % 10 == 0:
                    print(num_iter)

        return self.global_best.position, [log_likelihood_list, pos_list, vel_list]

    def _get_fitness(self, swarm):
        """Set fitness (probability) of the particles in swarm.

        :param swarm: PSO state
        :type swarm: list of Particle() instances of the swarm
        :return:
        :rtype:
        """
        position = [particle.position for particle in swarm]
        if self.pool is None:
            map_func = map
        else:
            map_func = self.pool.map
        ln_probability = list(map_func(self.func, position))

        for i, particle in enumerate(swarm):
            particle.fitness = ln_probability[i]
            particle.position = position[i]

    def _converged(self, it, p, m, n):
        """Check for convergence.

        :param it:
        :type it:
        :param p:
        :type p:
        :param m:
        :type m:
        :param n:
        :type n:
        :return:
        :rtype:
        """
        #        test = self._converged_space2(p=p)
        #        print(test)
        fit = self._converged_fit(it=it, p=p, m=m)
        if fit:
            space = self._converged_space(it=it, p=p, m=n)
            return space
        else:
            return False

    def _converged_fit(self, it, p, m):
        """

        :param it:
        :type it:
        :param p:
        :type p:
        :param m:
        :type m:
        :return:
        :rtype:
        """
        best_sort = np.sort(
            [particle.personal_best.fitness for particle in self.swarm]
        )[::-1]
        mean_fit = np.mean(best_sort[1 : int(math.floor(self.particleCount * p))])
        # print( "best %f, mean_fit %f, ration %f"%( self.global_best[0],
        # mean_fit, abs((self.global_best[0]-mean_fit))))
        return abs(self.global_best.fitness - mean_fit) < m

    def _converged_space(self, it, p, m):
        """

        :param it:
        :type it:
        :param p:
        :type p:
        :param m:
        :type m:
        :return:
        :rtype:
        """
        sorted_swarm = [particle for particle in self.swarm]
        sorted_swarm.sort()
        best_of_best = sorted_swarm[0 : int(floor(self.particleCount * p))]

        diffs = []
        for particle in best_of_best:
            diffs.append(
                np.array(self.global_best.position) - np.array(particle.position)
            )

        max_norm = max(list(map(np.linalg.norm, diffs)))
        return abs(max_norm) < m

    def _converged_space2(self, p):
        """

        :param p:
        :type p:
        :return:
        :rtype:
        """
        # Andres N. Ruiz et al.
        sorted_swarm = [particle for particle in self.swarm]
        sorted_swarm.sort()
        best_of_best = sorted_swarm[0 : int(floor(self.particleCount * p))]

        positions = [particle.position for particle in best_of_best]
        means = np.mean(positions, axis=0)
        delta = np.mean(
            (means - np.array(self.global_best.position))
            / np.array(self.global_best.position)
        )
        return np.log10(delta) < -3.0

    def is_master(self):
        """Check if the current processor is the master.

        :return:
        :rtype:
        """
        if self.pool is None:
            return True
        else:
            return self.pool.is_master()

    def _acceptable_convergence(self, chi_square_tolerance):
        chi_square = -2 * self.global_best.fitness

        if np.min(chi_square) < chi_square_tolerance:
            return True
        else:
            return False


class Particle(object):
    """Implementation of a single particle.

    :param position: the position of the particle in the parameter space
    :param velocity: the velocity of the particle
    :param fitness: the current fitness of the particle
    """

    def __init__(self, position, velocity, fitness=0):
        """

        :param position: parameter positions
        :param velocity: parameter velocity
        :param fitness:
        """
        self.position = [p for p in position]
        self.velocity = [v for v in velocity]

        self.fitness = fitness
        self.param_count = len(self.position)
        self._personal_best = None

    @property
    def personal_best(self):
        """

        :return:
        :rtype:
        """
        if self._personal_best is None:
            return self
        else:
            return self._personal_best

    @classmethod
    def create(cls, param_count):
        """Creates a new particle without position, velocity and -inf as fitness."""

        return Particle(
            np.array([[]] * param_count), np.array([[]] * param_count), -np.Inf
        )

    def update_personal_best(self):
        """Sets the current particle representation as personal best."""
        self._personal_best = self.copy()

    def copy(self):
        """Creates a copy of itself."""
        return Particle(copy(self.position), copy(self.velocity), self.fitness)

    def __str__(self):
        """Get a `str` object for the particle state.

        :return:
        :rtype:
        """
        return "{:f}, pos: {:s} velocity: {:s}".format(
            self.fitness, self.position, self.velocity
        )

    def __lt__(self, other):
        return self.fitness > other.fitness

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state

    def __unicode__(self):
        return self.__str__()


class _FunctionWrapper(object):
    """This is a hack to make the likelihood function pickleable when ``args`` or
    ``kwargs`` are also included.

    This hack is copied from
    emcee: https://github.com/dfm/emcee/.
    """

    def __init__(self, f, args, kwargs):
        self.f = f
        self.args = [] if args is None else args
        self.kwargs = {} if kwargs is None else kwargs

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__ = state

    def __call__(self, x):
        try:
            return self.f(x, *self.args, **self.kwargs)
        except:  # pragma: no cover
            import traceback

            print("PSO: Exception while calling your likelihood function:")
            print("  params:", x)
            print("  args:", self.args)
            print("  kwargs:", self.kwargs)
            print("  exception:")
            traceback.print_exc()
            raise



__author__ = "sibirrer"

import numpy as np
import numpy.testing as npt
from lenstronomy.ImSim.Numerics.adaptive_numerics import AdaptiveConvolution
from lenstronomy.ImSim.Numerics.convolution import SubgridKernelConvolution
from lenstronomy.LightModel.light_model import LightModel
import lenstronomy.Util.util as util
import pytest


class TestAdaptiveConvolution(object):
    def setup_method(self):
        self.supersampling_factor = 3
        lightModel = LightModel(light_model_list=["GAUSSIAN"])
        self.delta_pix = 1.0
        x, y = util.make_grid(20, deltapix=self.delta_pix)
        x_sub, y_sub = util.make_grid(
            20 * self.supersampling_factor,
            deltapix=self.delta_pix / self.supersampling_factor,
        )
        kwargs = [{"amp": 1, "sigma": 2, "center_x": 0, "center_y": 0}]
        flux = lightModel.surface_brightness(x, y, kwargs)
        self.model = util.array2image(flux)
        flux_sub = lightModel.surface_brightness(x_sub, y_sub, kwargs)
        self.model_sub = util.array2image(flux_sub)

        x, y = util.make_grid(5, deltapix=self.delta_pix)
        kwargs_kernel = [{"amp": 1, "sigma": 1, "center_x": 0, "center_y": 0}]
        kernel = lightModel.surface_brightness(x, y, kwargs_kernel)
        self.kernel = util.array2image(kernel) / np.sum(kernel)

        x_sub, y_sub = util.make_grid(
            5 * self.supersampling_factor,
            deltapix=self.delta_pix / self.supersampling_factor,
        )
        kernel_sub = lightModel.surface_brightness(x_sub, y_sub, kwargs_kernel)
        self.kernel_sub = util.array2image(kernel_sub) / np.sum(kernel_sub)

    def test_convolve2d(self):
        # kernel_supersampled = kernel_util.subgrid_kernel(self.kernel, self.supersampling_factor, odd=True, num_iter=5)
        subgrid_conv = SubgridKernelConvolution(
            self.kernel_sub,
            self.supersampling_factor,
            supersampling_kernel_size=None,
            convolution_type="fft",
        )
        model_subgrid_conv = subgrid_conv.convolution2d(self.model_sub)

        conv_supersample_pixels = np.zeros_like(self.model)
        conv_supersample_pixels = np.array(conv_supersample_pixels, dtype=bool)
        conv_supersample_pixels[self.model > np.max(self.model) / 20] = True
        adaptive_conv = AdaptiveConvolution(
            self.kernel_sub,
            self.supersampling_factor,
            conv_supersample_pixels,
            supersampling_kernel_size=5,
            compute_pixels=None,
        )

        model_adaptive_conv = adaptive_conv.convolve2d(self.model_sub)
        npt.assert_almost_equal(
            np.sum(model_subgrid_conv), np.sum(model_adaptive_conv), decimal=2
        )
        npt.assert_almost_equal(model_subgrid_conv, model_adaptive_conv, decimal=2)

        conv_supersample_pixels = np.zeros_like(self.model)
        conv_supersample_pixels = np.array(conv_supersample_pixels, dtype=bool)
        conv_supersample_pixels[self.model > np.max(self.model) / 2] = True
        adaptive_conv = AdaptiveConvolution(
            self.kernel_sub,
            self.supersampling_factor,
            conv_supersample_pixels,
            supersampling_kernel_size=1,
            compute_pixels=None,
        )

        model_adaptive_conv = adaptive_conv.convolve2d(self.model_sub)

        npt.assert_almost_equal(
            np.sum(model_subgrid_conv), np.sum(model_adaptive_conv), decimal=2
        )
        npt.assert_almost_equal(model_subgrid_conv, model_adaptive_conv, decimal=2)


if __name__ == "__main__":
    pytest.main()



"""Tests for `galkin` module."""
import pytest
import numpy.testing as npt
import numpy as np
import unittest

from lenstronomy.GalKin.anisotropy import Anisotropy


class TestAnisotropy(object):
    def setup_method(self):
        self._r_array = np.array([2.0, 3.0])
        self._R_array = 1.0

    def test_K(self):
        anisoClass = Anisotropy(anisotropy_type="const")
        kwargs = {"beta": 1.0}
        k = anisoClass.K(self._r_array, self._R_array, **kwargs)
        npt.assert_almost_equal(k[0], 0.61418484930437822, decimal=5)

        kwargs = {"beta": -0.49}
        k = anisoClass.K(self._r_array, self._R_array, **kwargs)
        npt.assert_almost_equal(k[0], 0.7645553632433857, decimal=5)

        anisoClass = Anisotropy(anisotropy_type="Colin")
        kwargs = {"r_ani": 1}
        k = anisoClass.K(self._r_array, self._R_array, **kwargs)
        npt.assert_almost_equal(k[0], 0.91696135187291117, decimal=5)
        k = anisoClass.K(self._r_array, self._R_array - 0.001, **kwargs)
        npt.assert_almost_equal(k[0], 0.91696135187291117, decimal=2)
        k = anisoClass.K(self._r_array, self._R_array + 0.001, **kwargs)
        npt.assert_almost_equal(k[0], 0.91696135187291117, decimal=2)

        anisoClass = Anisotropy(anisotropy_type="radial")
        kwargs = {}
        k = anisoClass.K(self._r_array, self._R_array, **kwargs)
        npt.assert_almost_equal(k[0], 0.61418484930437856, decimal=5)

        anisoClass = Anisotropy(anisotropy_type="isotropic")
        kwargs = {}
        k = anisoClass.K(self._r_array, self._R_array, **kwargs)
        npt.assert_almost_equal(k[0], 0.8660254037844386, decimal=5)

        anisoClass = Anisotropy(anisotropy_type="OM")
        kwargs = {"r_ani": 1}
        k = anisoClass.K(self._r_array, self._R_array, **kwargs)
        npt.assert_almost_equal(k[0], 0.95827704196894481, decimal=5)

        R = 0.1
        k = anisoClass.K(R, R, **kwargs)
        npt.assert_almost_equal(k, 0, decimal=5)

    def test_beta(self):
        r = 2.0

        anisoClass = Anisotropy(anisotropy_type="const")
        kwargs = {"beta": 1.0}
        beta = anisoClass.beta_r(r, **kwargs)
        npt.assert_almost_equal(beta, 1, decimal=5)

        anisoClass = Anisotropy(anisotropy_type="Colin")
        kwargs = {"r_ani": 1}
        beta = anisoClass.beta_r(r, **kwargs)
        npt.assert_almost_equal(beta, 1.0 / 3, decimal=5)

        anisoClass = Anisotropy(anisotropy_type="radial")
        kwargs = {}
        beta = anisoClass.beta_r(r, **kwargs)
        npt.assert_almost_equal(beta, 1, decimal=5)

        anisoClass = Anisotropy(anisotropy_type="isotropic")
        kwargs = {}
        beta = anisoClass.beta_r(r, **kwargs)
        npt.assert_almost_equal(beta, 0, decimal=5)

        anisoClass = Anisotropy(anisotropy_type="OM")
        kwargs = {"r_ani": 1}
        beta = anisoClass.beta_r(r, **kwargs)
        npt.assert_almost_equal(beta, 0.8, decimal=5)

    def test_radial_anisotropy(self):
        # radial
        r = 2.0
        R = 1.0
        radial = Anisotropy(anisotropy_type="radial")
        kwargs_rad = {}
        beta = radial.beta_r(r, **kwargs_rad)
        k = radial.K(r, R, **kwargs_rad)
        f = radial.anisotropy_solution(r, **kwargs_rad)
        assert f == r**2
        const = Anisotropy(anisotropy_type="const")
        kwargs = {"beta": beta}
        print(beta, "beta")
        # kwargs = {'beta': 1}
        k_mamon = const.K(r, R, **kwargs)
        print(k, k_mamon)
        npt.assert_almost_equal(k, k_mamon, decimal=5)

    def test_isotropic_anisotropy(self):
        # radial
        r = 2.0
        R = 1.0
        isotropic = Anisotropy(anisotropy_type="isotropic")
        kwargs_iso = {}
        beta = isotropic.beta_r(r, **kwargs_iso)
        k = isotropic.K(r, R, **kwargs_iso)
        f = isotropic.anisotropy_solution(r, **kwargs_iso)
        assert f == 1
        print(beta, "test")
        const = Anisotropy(anisotropy_type="const")
        kwargs = {"beta": beta}
        k_const = const.K(r, R, **kwargs)
        print(k, k_const)
        npt.assert_almost_equal(k, k_const, decimal=5)

    def test_generalizedOM(self):
        # generalized OM model
        gom = Anisotropy(anisotropy_type="GOM")
        r = self._r_array
        R = 2
        om = Anisotropy(anisotropy_type="OM")
        kwargs_om = {"r_ani": 1.0}
        kwargs_gom = {"r_ani": 1.0, "beta_inf": 1.0}
        beta_gom = gom.beta_r(r, **kwargs_gom)
        beta_om = om.beta_r(r, **kwargs_om)
        npt.assert_almost_equal(beta_gom, beta_om, decimal=5)

        f_om = om.anisotropy_solution(r, **kwargs_om)
        f_gom = gom.anisotropy_solution(r, **kwargs_gom)
        npt.assert_almost_equal(f_gom, f_om, decimal=5)

        K_gom = gom.K(r, R, **kwargs_gom)
        K_om = om.K(r, R, **kwargs_om)
        npt.assert_almost_equal(K_gom, K_om, decimal=3)
        assert hasattr(gom._model, "_f_12_interp")
        assert hasattr(gom._model, "_f_32_interp")
        gom.delete_anisotropy_cache()
        if hasattr(gom._model, "_f_12_interp"):
            assert False
        if hasattr(gom._model, "_f_32_interp"):
            assert False

        from lenstronomy.GalKin.anisotropy import GeneralizedOM

        gom_class = GeneralizedOM()
        _F = gom_class._F(a=3 / 2.0, z=0.5, beta_inf=1)
        _F_array = gom_class._F(a=3 / 2.0, z=np.array([0.5]), beta_inf=1)
        npt.assert_almost_equal(_F_array[0], _F, decimal=5)


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            Anisotropy(anisotropy_type="wrong")
        with self.assertRaises(ValueError):
            ani = Anisotropy(anisotropy_type="Colin")
            ani.K(r=1, R=2, r_ani=1)

        with self.assertRaises(ValueError):
            ani = Anisotropy(anisotropy_type="const")
            ani.anisotropy_solution(r=1)

        with self.assertRaises(ValueError):
            const = Anisotropy(anisotropy_type="const")
            kwargs = {"beta": 1}
            f_const = const.anisotropy_solution(r=1, **kwargs)


if __name__ == "__main__":
    pytest.main()



from lenstronomy.LensModel.Solver.lens_equation_solver import LensEquationSolver
import numpy as np

__all__ = ["PSBase", "_expand_to_array", "_shrink_array"]


class PSBase(object):
    """Base point source type class."""

    def __init__(
        self,
        lens_model=None,
        fixed_magnification=False,
        additional_images=False,
        index_lens_model_list=None,
        point_source_frame_list=None,
    ):
        """

        :param lens_model: instance of the LensModel() class
        :param fixed_magnification: bool. If True, magnification
         ratio of point sources is fixed to the one given by the lens model
        :param additional_images: bool. If True, search for additional images of the same source is conducted.
        :param index_lens_model_list: list (length of different patches/bands) of integer lists, e.g., [[0, 1], [2, 3]];
         evaluating a subset of the lens models per individual bands. If this keyword is set, the image positions need
         to have a specified band/frame assigned to it
        :param point_source_frame_list: list of lists mirroring the structure of the image positions.
         Integers correspond to the i'th list entry of index_lens_model_list indicating in which frame/band the image is
         appearing
        """
        self._lens_model = lens_model
        if index_lens_model_list is not None:
            k_list = []
            for point_source_frame in point_source_frame_list:
                k_list.append(index_lens_model_list[point_source_frame])
            self.k_list = k_list
        else:
            self.k_list = None
        if self._lens_model is None:
            self._solver = None
        else:
            self._solver = LensEquationSolver(lens_model)
        self._fixed_magnification = fixed_magnification
        self.additional_images = additional_images
        if fixed_magnification is True and additional_images is True:
            Warning(
                "The combination of fixed_magnification=True and additional_image=True is not optimal for the "
                "current computation. If you see this warning, please approach the developers."
            )

    def image_position(self, kwargs_ps, **kwargs):
        """On-sky position.

        :param kwargs_ps: keyword argument of point source model
        :return: numpy array of x, y image positions
        """
        raise ValueError(
            "image_position definition is not defined in the profile you want to execute."
        )

    def source_position(self, kwargs_ps, **kwargs):
        """Original unlensed position.

        :param kwargs_ps: keyword argument of point source model
        :return: numpy array of x, y source positions
        """
        raise ValueError(
            "source_position definition is not defined in the profile you want to execute."
        )

    def image_amplitude(self, kwargs_ps, *args, **kwargs):
        """Amplitudes as observed on the sky.

        :param kwargs_ps: keyword argument of point source model
        :param kwargs: keyword arguments of function call
        :return: numpy array of amplitudes
        """
        raise ValueError(
            "source_position definition is not defined in the profile you want to execute."
        )

    def source_amplitude(self, kwargs_ps, **kwargs):
        """Intrinsic source amplitudes (without lensing magnification, but still
        apparent)

        :param kwargs_ps: keyword argument of point source model
        :param kwargs: keyword arguments of function call (which are not used for this
            object
        :return: numpy array of amplitudes
        """
        raise ValueError(
            "source_position definition is not defined in the profile you want to execute."
        )

    def update_lens_model(self, lens_model_class):
        """Update LensModel() and LensEquationSolver() instance.

        :param lens_model_class: LensModel() class instance
        :return: internal lensModel class updated
        """
        self._lens_model = lens_model_class
        if lens_model_class is None:
            self._solver = None
        else:
            self._solver = LensEquationSolver(lens_model_class)


def _expand_to_array(array, num):
    """

    :param array: float/int or numpy array
    :param num: number of array entries expected in array
    :return: array of size num
    """
    if np.isscalar(array):
        return np.ones(num) * array
    elif len(array) < num:
        out = np.zeros(num)
        out[0 : len(array)] = array
        return out
    else:
        return array


def _shrink_array(array, num):
    """
    :param array: float/int or numpy array
    :param num: number of array entries expected in array
    :return: array of size num, or scalar if array is a scalar
    """
    if np.isscalar(array):
        return array
    elif len(array) > num:
        array_return = array[:num]
        return array_return
    elif len(array) < num:
        raise ValueError(
            "the length of the array (%s) needs to be larger or equal than the designated length %s "
            % (len(array), num)
        )
    else:
        return array



__author__ = "sibirrer"

import pytest
import numpy as np
import os
import lenstronomy.Util.simulation_util as sim_util
from lenstronomy.ImSim.image_model import ImageModel
from lenstronomy.Sampling.likelihood import LikelihoodModule
from lenstronomy.Sampling.parameters import Param
from lenstronomy.LensModel.lens_model import LensModel
from lenstronomy.LightModel.light_model import LightModel
from lenstronomy.Sampling.sampler import Sampler, choose_pool
from lenstronomy.Data.imaging_data import ImageData
from lenstronomy.Data.psf import PSF


class TestSampler(object):
    """Test the fitting sequences."""

    def setup_method(self):
        # data specifics
        sigma_bkg = 0.05  # background noise per pixel
        exp_time = 100  # exposure time (arbitrary units, flux per pixel is in units #photons/exp_time unit)
        numPix = 10  # cutout pixel size
        deltaPix = 0.1  # pixel size in arcsec (area per pixel = deltaPix**2)
        fwhm = 0.5  # full width half max of PSF

        # PSF specification

        kwargs_data = sim_util.data_configure_simple(
            numPix, deltaPix, exp_time, sigma_bkg
        )
        data_class = ImageData(**kwargs_data)
        kwargs_psf_gaussian = {
            "psf_type": "GAUSSIAN",
            "fwhm": fwhm,
            "pixel_size": deltaPix,
        }
        psf = PSF(**kwargs_psf_gaussian)
        kwargs_psf = {
            "psf_type": "PIXEL",
            "kernel_point_source": psf.kernel_point_source,
        }
        psf_class = PSF(**kwargs_psf)
        kwargs_spemd = {
            "theta_E": 1.0,
            "gamma": 1.8,
            "center_x": 0,
            "center_y": 0,
            "e1": 0.1,
            "e2": 0.1,
        }

        lens_model_list = ["SPEP"]
        self.kwargs_lens = [kwargs_spemd]
        lens_model_class = LensModel(lens_model_list=lens_model_list)
        kwargs_sersic = {
            "amp": 1.0,
            "R_sersic": 0.1,
            "n_sersic": 2,
            "center_x": 0,
            "center_y": 0,
        }
        # 'SERSIC_ELLIPSE': elliptical Sersic profile
        kwargs_sersic_ellipse = {
            "amp": 1.0,
            "R_sersic": 0.6,
            "n_sersic": 3,
            "center_x": 0,
            "center_y": 0,
            "e1": 0.1,
            "e2": 0.1,
        }

        lens_light_model_list = ["SERSIC"]
        self.kwargs_lens_light = [kwargs_sersic]
        lens_light_model_class = LightModel(light_model_list=lens_light_model_list)
        source_model_list = ["SERSIC_ELLIPSE"]
        self.kwargs_source = [kwargs_sersic_ellipse]
        source_model_class = LightModel(light_model_list=source_model_list)

        kwargs_numerics = {
            "supersampling_factor": 1,
            "supersampling_convolution": False,
            "compute_mode": "regular",
        }
        imageModel = ImageModel(
            data_class,
            psf_class,
            lens_model_class,
            source_model_class,
            lens_light_model_class,
            kwargs_numerics=kwargs_numerics,
        )
        image_sim = sim_util.simulate_simple(
            imageModel, self.kwargs_lens, self.kwargs_source, self.kwargs_lens_light
        )

        data_class.update_data(image_sim)
        kwargs_data["image_data"] = image_sim
        kwargs_data_joint = {
            "multi_band_list": [[kwargs_data, kwargs_psf, kwargs_numerics]],
            "multi_band_type": "single-band",
        }
        self.data_class = data_class
        self.psf_class = psf_class

        kwargs_model = {
            "lens_model_list": lens_model_list,
            "source_light_model_list": source_model_list,
            "lens_light_model_list": lens_light_model_list,
            "fixed_magnification_list": [False],
        }
        self.kwargs_numerics = {"subgrid_res": 1, "psf_subgrid": False}

        kwargs_constraints = {
            "image_plane_source_list": [False] * len(source_model_list)
        }

        kwargs_likelihood = {
            "source_marg": False,
            "image_position_uncertainty": 0.004,
            "check_matched_source_position": False,
            "source_position_tolerance": 0.001,
            "source_position_sigma": 0.001,
        }
        self.param_class = Param(kwargs_model, **kwargs_constraints)
        self.Likelihood = LikelihoodModule(
            kwargs_data_joint=kwargs_data_joint,
            kwargs_model=kwargs_model,
            param_class=self.param_class,
            **kwargs_likelihood
        )
        self.sampler = Sampler(likelihoodModule=self.Likelihood)

    def test_pso(self):
        n_particles = 2
        n_iterations = 2
        result, chain = self.sampler.pso(
            n_particles,
            n_iterations,
            lower_start=None,
            upper_start=None,
            threadCount=1,
            init_pos=None,
            mpi=False,
            print_key="PSO",
        )

        assert len(result) == 16

    def test_mcmc_emcee(self):
        n_walkers = 36
        n_run = 2
        n_burn = 2
        mean_start = self.param_class.kwargs2args(
            kwargs_lens=self.kwargs_lens,
            kwargs_source=self.kwargs_source,
            kwargs_lens_light=self.kwargs_lens_light,
        )
        sigma_start = np.ones_like(mean_start) * 0.1
        samples, dist = self.sampler.mcmc_emcee(
            n_walkers, n_run, n_burn, mean_start, sigma_start, mpi=False
        )
        assert len(samples) == n_walkers * n_run
        assert len(dist) == len(samples)

        # test of backup file
        # 1) run a chain specifiying a backup file name
        backup_filename = "test_mcmc_emcee.h5"
        samples_1, dist_1 = self.sampler.mcmc_emcee(
            n_walkers,
            n_run,
            n_burn,
            mean_start,
            sigma_start,
            mpi=False,
            backend_filename=backup_filename,
        )
        assert len(samples_1) == n_walkers * n_run
        # 2) run a chain starting from the backup of previous run
        samples_2, dist_2 = self.sampler.mcmc_emcee(
            n_walkers,
            n_run,
            n_burn,
            mean_start,
            sigma_start,
            mpi=False,
            backend_filename=backup_filename,
            start_from_backend=True,
        )
        assert len(samples_2) == len(samples_1) + n_walkers * n_run
        assert len(dist_2) == len(samples_2)

        os.remove(backup_filename)  # just remove the backup file created above

    def test_mcmc_zeus(self):
        n_walkers = 36
        n_run = 2
        n_burn = 2
        mean_start = self.param_class.kwargs2args(
            kwargs_lens=self.kwargs_lens,
            kwargs_source=self.kwargs_source,
            kwargs_lens_light=self.kwargs_lens_light,
        )
        sigma_start = np.ones_like(mean_start) * 0.1
        samples, dist = self.sampler.mcmc_zeus(
            n_walkers, n_run, n_burn, mean_start, sigma_start
        )
        assert len(samples) == n_walkers * n_run
        assert len(dist) == len(samples)

        # test of backup file
        backup_filename = "test_mcmc_zeus.h5"
        samples_1, dist_1 = self.sampler.mcmc_zeus(
            n_walkers,
            n_run,
            n_burn,
            mean_start,
            sigma_start,
            backend_filename=backup_filename,
        )
        assert len(samples_1) == n_walkers * n_run

        os.remove(backup_filename)

        # test callbacks
        autocorrelation_callback = True
        splitr_callback = True
        miniter_callback = True

        samples_ac, dist_ac = self.sampler.mcmc_zeus(
            n_walkers,
            n_run,
            n_burn,
            mean_start,
            sigma_start,
            autocorrelation_callback=autocorrelation_callback,
        )
        assert len(samples_ac) == n_walkers * n_run

        # the default nsplits is 2, here we set it to 1 because the test chain is very short
        # i.e. 4/3 != integer; 4/2 = integer
        samples_sp, dist_sp = self.sampler.mcmc_zeus(
            n_walkers,
            n_run,
            n_burn,
            mean_start,
            sigma_start,
            splitr_callback=splitr_callback,
            nsplits=1,
        )
        assert len(samples_sp) == n_walkers * n_run

        samples_mi, dist_mi = self.sampler.mcmc_zeus(
            n_walkers,
            n_run,
            n_burn,
            mean_start,
            sigma_start,
            miniter_callback=miniter_callback,
        )
        assert len(samples_mi) == n_walkers * n_run


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
from lenstronomy.LensModel.Profiles.nfw import NFW
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["NFW_ELLIPSE"]


class NFW_ELLIPSE(LensProfileBase):
    """This class contains functions concerning the NFW profile with an ellipticity
    defined in the potential parameterization of alpha_Rs and Rs is the same as for the
    spherical NFW profile.

    from Glose & Kneib: https://cds.cern.ch/record/529584/files/0112138.pdf

    relation are: R_200 = c * Rs
    """

    profile_name = "NFW_ELLIPSE"
    param_names = ["Rs", "alpha_Rs", "e1", "e2", "center_x", "center_y"]
    lower_limit_default = {
        "Rs": 0,
        "alpha_Rs": 0,
        "e1": -0.5,
        "e2": -0.5,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "Rs": 100,
        "alpha_Rs": 10,
        "e1": 0.5,
        "e2": 0.5,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self, interpol=False, num_interp_X=1000, max_interp_X=10):
        """

        :param interpol: bool, if True, interpolates the functions F(), g() and h()
        :param num_interp_X: int (only considered if interpol=True), number of interpolation elements in units of r/r_s
        :param max_interp_X: float (only considered if interpol=True), maximum r/r_s value to be interpolated
         (returning zeros outside)
        """
        self.nfw = NFW(
            interpol=interpol, num_interp_X=num_interp_X, max_interp_X=max_interp_X
        )
        self._diff = 0.0000000001
        super(NFW_ELLIPSE, self).__init__()

    def function(self, x, y, Rs, alpha_Rs, e1, e2, center_x=0, center_y=0):
        """Returns elliptically distorted NFW lensing potential.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param alpha_Rs: deflection (angular units) at projected Rs
        :param e1: eccentricity component in x-direction
        :param e2: eccentricity component in y-direction
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: lensing potential
        """
        x_, y_ = param_util.transform_e1e2_square_average(
            x, y, e1, e2, center_x, center_y
        )
        R_ = np.sqrt(x_**2 + y_**2)
        rho0_input = self.nfw.alpha2rho0(alpha_Rs=alpha_Rs, Rs=Rs)
        if Rs < 0.0000001:
            Rs = 0.0000001
        f_ = self.nfw.nfwPot(R_, Rs, rho0_input)
        return f_

    def derivatives(self, x, y, Rs, alpha_Rs, e1, e2, center_x=0, center_y=0):
        """Returns df/dx and df/dy of the function, calculated as an elliptically
        distorted deflection angle of the spherical NFW profile.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param alpha_Rs: deflection (angular units) at projected Rs
        :param e1: eccentricity component in x-direction
        :param e2: eccentricity component in y-direction
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: deflection in x-direction, deflection in y-direction
        """
        x_, y_ = param_util.transform_e1e2_square_average(
            x, y, e1, e2, center_x, center_y
        )
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        cos_phi = np.cos(phi_G)
        sin_phi = np.sin(phi_G)
        e = param_util.q2e(q)
        # e = abs(1 - q)
        R_ = np.sqrt(x_**2 + y_**2)
        rho0_input = self.nfw.alpha2rho0(alpha_Rs=alpha_Rs, Rs=Rs)
        if Rs < 0.0000001:
            Rs = 0.0000001
        f_x_prim, f_y_prim = self.nfw.nfwAlpha(R_, Rs, rho0_input, x_, y_)
        f_x_prim *= np.sqrt(1 - e)
        f_y_prim *= np.sqrt(1 + e)
        f_x = cos_phi * f_x_prim - sin_phi * f_y_prim
        f_y = sin_phi * f_x_prim + cos_phi * f_y_prim
        return f_x, f_y

    def hessian(self, x, y, Rs, alpha_Rs, e1, e2, center_x=0, center_y=0):
        """Returns Hessian matrix of function d^2f/dx^2, d^f/dy^2, d^2/dxdy the
        calculation is performed as a numerical differential from the deflection field.
        Analytical relations are possible.

        :param x: angular position (normally in units of arc seconds)
        :param y: angular position (normally in units of arc seconds)
        :param Rs: turn over point in the slope of the NFW profile in angular unit
        :param alpha_Rs: deflection (angular units) at projected Rs
        :param e1: eccentricity component in x-direction
        :param e2: eccentricity component in y-direction
        :param center_x: center of halo (in angular units)
        :param center_y: center of halo (in angular units)
        :return: d^2f/dx^2, d^2/dxdy, d^2/dydx, d^f/dy^2
        """
        alpha_ra, alpha_dec = self.derivatives(
            x, y, Rs, alpha_Rs, e1, e2, center_x, center_y
        )
        diff = self._diff
        alpha_ra_dx, alpha_dec_dx = self.derivatives(
            x + diff, y, Rs, alpha_Rs, e1, e2, center_x, center_y
        )
        alpha_ra_dy, alpha_dec_dy = self.derivatives(
            x, y + diff, Rs, alpha_Rs, e1, e2, center_x, center_y
        )

        f_xx = (alpha_ra_dx - alpha_ra) / diff
        f_xy = (alpha_ra_dy - alpha_ra) / diff
        f_yx = (alpha_dec_dx - alpha_dec) / diff
        f_yy = (alpha_dec_dy - alpha_dec) / diff

        return f_xx, f_xy, f_yx, f_yy

    def mass_3d_lens(self, r, Rs, alpha_Rs, e1=1, e2=0):
        """

        :param r: radius (in angular units)
        :param Rs:
        :param alpha_Rs:
        :param e1:
        :param e2:
        :return:
        """
        return self.nfw.mass_3d_lens(r, Rs, alpha_Rs)

    def density_lens(self, r, Rs, alpha_Rs, e1=1, e2=0):
        """Computes the density at 3d radius r given lens model parameterization. The
        integral in the LOS projection of this quantity results in the convergence
        quantity.

        :param r: 3d radios
        :param Rs: turn-over radius of NFW profile
        :param alpha_Rs: deflection at Rs
        :return: density rho(r)
        """
        return self.nfw.density_lens(r, Rs, alpha_Rs)



"""This module provides helper classes for managing sample parameters.

This is for internal use, if you are not modifying lenstronomy sampling to include new
parameters you can safely ignore this.
"""

__author__ = "jhodonnell"
__all__ = ["ModelParamGroup", "SingleParam", "ArrayParam"]

import numpy as np


class ModelParamGroup:
    """This abstract class represents any lenstronomy fitting parameters used in the
    Param class.

    Subclasses should implement num_params(), set_params(), and get_params() to convert
    parameters from lenstronomy's semantic dictionary format to a flattened array format
    and back.

    This class also contains three static methods to easily aggregate groups of
    parameter classes, called `compose_num_params()`, `compose_set_params()`, and
    `compose_get_params()`.
    """

    def num_params(self):
        """Tells the number of parameters that this group samples and their names.

        :returns: 2-tuple of (num param, list of names)
        """
        raise NotImplementedError

    def set_params(self, kwargs):
        """Converts lenstronomy semantic parameters in dictionary format into a
        flattened array of parameters.

        The flattened array is for use in optimization algorithms, e.g. MCMC, Particle
        swarm, etc.

        :returns: flattened array of parameters as floats
        """
        raise NotImplementedError

    def get_params(self, args, i):
        """Converts a flattened array of parameters back into a lenstronomy dictionary,
        starting at index i.

        :param args: flattened arguments to convert to lenstronomy format
        :type args: list
        :param i: index to begin at in args
        :type i: int
        :returns: dictionary of parameters
        """
        raise NotImplementedError

    @staticmethod
    def compose_num_params(each_group, *args, **kwargs):
        """Aggregates the number of parameters for a group of parameter groups, calling
        each instance's `num_params()` method and combining the results.

        :param each_group: collection of parameter groups. Should each be subclasses of
            ModelParamGroup.
        :type each_group: list
        :param args: Extra arguments to be passed to each call of `num_params()`
        :param kwargs: Extra keyword arguments to be passed to each call of
            `num_params()`
        :returns: As in each individual `num_params()`, a 2-tuple of (num params, list
            of param names)
        """
        tot_param = 0
        param_names = []
        for group in each_group:
            npar, names = group.num_params(*args, **kwargs)
            tot_param += npar
            param_names += names
        return tot_param, param_names

    @staticmethod
    def compose_set_params(each_group, param_kwargs, *args, **kwargs):
        """Converts lenstronomy semantic arguments in dictionary format to a flattened
        list of floats for use in optimization/fitting algorithms. Combines the results
        for a set of arbitrarily many parameter groups.

        :param each_group: collection of parameter groups. Should each be subclasses of
            ModelParamGroup.
        :type each_group: list
        :param param_kwargs: the kwargs to process
        :type param_kwargs: dict
        :param args: Extra arguments to be passed to each call of `set_params()`
        :param kwargs: Extra keyword arguments to be passed to each call of
            `set_params()`
        :returns: As in each individual `set_params()`, a list of floats
        """
        output_args = []
        for group in each_group:
            output_args += group.set_params(param_kwargs, *args, **kwargs)
        return output_args

    @staticmethod
    def compose_get_params(each_group, flat_args, i, *args, **kwargs):
        """Converts a flattened array of parameters to lenstronomy semantic parameters
        in dictionary format. Combines the results for a set of arbitrarily many
        parameter groups.

        :param each_group: collection of parameter groups. Should each be subclasses of ModelParamGroup.
        :type each_group: list
        :param flat_args: the input array of parameters
        :type flat_args: list
        :param i: the index in `flat_args` to start at
        :type i: int
        :param args: Extra arguments to be passed to each call of `set_params()`
        :param kwargs: Extra keyword arguments to be passed to each call of `set_params()`

        :returns: As in each individual `get_params()`, a 2-tuple of (dictionary of params, new index)
        """
        output_kwargs = {}
        for group in each_group:
            kwargs_grp, i = group.get_params(flat_args, i, *args, **kwargs)
            output_kwargs = dict(output_kwargs, **kwargs_grp)
        return output_kwargs, i


class SingleParam(ModelParamGroup):
    """Helper for handling parameters which are a single float.

    Subclasses should define:

    :param on: Whether this parameter is sampled
    :type on: bool
    :param param_names: List of strings, the name of each parameter
    :param _kwargs_lower: Dictionary. Lower bounds of each parameter
    :param _kwargs_upper: Dictionary. Upper bounds of each parameter
    """

    def __init__(self, on):
        """
        :param on: Whether this paramter should be sampled
        :type on: bool
        """
        self._on = bool(on)

    def num_params(self, kwargs_fixed):
        """Tells the number of parameters that this group samples and their names.

        :param kwargs_fixed: Dictionary of fixed arguments
        :type kwargs_fixed: dict
        :returns: 2-tuple of (num param, list of names)
        """
        if self.on:
            npar, names = 0, []
            for name in self.param_names:
                if name not in kwargs_fixed:
                    npar += 1
                    names.append(name)
            return npar, names
        return 0, []

    def set_params(self, kwargs, kwargs_fixed):
        """Converts lenstronomy semantic parameters in dictionary format into a
        flattened array of parameters.

        The flattened array is for use in optimization algorithms, e.g. MCMC, Particle
        swarm, etc.

        :param kwargs: lenstronomy parameters to flatten
        :type kwargs: dict
        :param kwargs_fixed: Dictionary of fixed arguments
        :type kwargs_fixed: dict
        :returns: flattened array of parameters as floats
        """
        if self.on:
            output = []
            for name in self.param_names:
                if name not in kwargs_fixed:
                    output.append(kwargs[name])
            return output
        return []

    def get_params(self, args, i, kwargs_fixed, kwargs_upper=None, kwargs_lower=None):
        """Converts a flattened array of parameters back into a lenstronomy dictionary,
        starting at index i.

        :param args: flattened arguments to convert to lenstronomy format
        :type args: list
        :param i: index to begin at in args
        :type i: int
        :param kwargs_fixed: Dictionary of fixed arguments
        :type kwargs_fixed: dict
        :returns: dictionary of parameters
        """
        out = {}
        if self.on:
            for name in self.param_names:
                if name in kwargs_fixed:
                    out[name] = kwargs_fixed[name]
                else:
                    out[name] = args[i]
                    if kwargs_lower is not None:
                        out[name] = np.maximum(out[name], kwargs_lower[name])
                    if kwargs_upper is not None:
                        out[name] = np.minimum(out[name], kwargs_upper[name])
                    i += 1
        return out, i

    @property
    def kwargs_lower(self):
        if not self.on:
            return {}
        return self._kwargs_lower

    @property
    def kwargs_upper(self):
        if not self.on:
            return {}
        return self._kwargs_upper

    @property
    def on(self):
        return self._on


class ArrayParam(ModelParamGroup):
    """Helper for handling parameters which are an array of values. Examples include
    mass_scaling, which is an array of scaling parameters, and wavelet or gaussian
    decompositions which have different coefficients for each mode.

    Subclasses should define:

    :param on:  Whether this parameter is sampled
    :type on: bool
    :param param_names: Dictionary mapping the name of each parameter to the number of values needed.
    :param _kwargs_lower: Dictionary. Lower bounds of each parameter
    :param _kwargs_upper: Dictionary. Upper bounds of each parameter
    """

    def __init__(self, on):
        """
        :param on: Whether this paramter should be sampled
        :type on: bool
        """
        self._on = bool(on)

    def num_params(self, kwargs_fixed):
        """Tells the number of parameters that this group samples and their names.

        :param kwargs_fixed: Dictionary of fixed arguments
        :type kwargs_fixed: dict
        :returns: 2-tuple of (num param, list of names)
        """
        if not self.on:
            return 0, []

        npar = 0
        names = []
        for name, count in self.param_names.items():
            if name not in kwargs_fixed:
                npar += count
                names += [name] * count

        return npar, names

    def set_params(self, kwargs, kwargs_fixed):
        """Converts lenstronomy semantic parameters in dictionary format into a
        flattened array of parameters.

        The flattened array is for use in optimization algorithms, e.g. MCMC, Particle
        swarm, etc.

        :param kwargs: lenstronomy parameters to flatten
        :type kwargs: dict
        :param kwargs_fixed: Dictionary of fixed arguments
        :type kwargs_fixed: dict
        :returns: flattened array of parameters as floats
        """
        if not self.on:
            return []

        args = []
        for name, count in self.param_names.items():
            if name not in kwargs_fixed:
                args.extend(kwargs[name])
        return args

    def get_params(self, args, i, kwargs_fixed, kwargs_lower=None, kwargs_upper=None):
        """Converts a flattened array of parameters back into a lenstronomy dictionary,
        starting at index i.

        :param args: flattened arguments to convert to lenstronomy format
        :type args: list
        :param i: index to begin at in args
        :type i: int
        :param kwargs_fixed: Dictionary of fixed arguments
        :type kwargs_fixed: dict
        :param kwargs_lower: Dictionary of lower bounds
        :type kwargs_lower: dict
        :param kwargs_upper: Dictionary of upper bounds
        :type kwargs_upper: dict
        :returns: dictionary of parameters
        """
        if not self.on:
            return {}, i

        params = {}
        for name, count in self.param_names.items():
            if name not in kwargs_fixed:
                params[name] = args[i : i + count]

                if kwargs_lower is not None:
                    for j in range(len(params[name])):
                        if params[name][j] < kwargs_lower[name][j]:
                            params[name][j] = kwargs_lower[name][j]

                if kwargs_upper is not None:
                    for j in range(len(params[name])):
                        if params[name][j] > kwargs_upper[name][j]:
                            params[name][j] = kwargs_upper[name][j]

                i += count
            else:
                params[name] = kwargs_fixed[name]

        return params, i

    @property
    def kwargs_lower(self):
        if not self.on:
            return {}

        out = {}
        for name, count in self.param_names.items():
            out[name] = [self._kwargs_lower[name]] * count
        return out

    @property
    def kwargs_upper(self):
        if not self.on:
            return {}

        out = {}
        for name, count in self.param_names.items():
            out[name] = [self._kwargs_upper[name]] * count
        return out

    @property
    def on(self):
        return self._on



__author__ = "sibirrer"

# this file contains a class which describes the surface brightness of the light models

import numpy as np
from lenstronomy.Util.util import convert_bool_list
from lenstronomy.Conf import config_loader

convention_conf = config_loader.conventions_conf()
sersic_major_axis_conf = convention_conf.get("sersic_major_axis", False)

__all__ = ["LightModelBase"]


_MODELS_SUPPORTED = [
    "GAUSSIAN",
    "GAUSSIAN_ELLIPSE",
    "ELLIPSOID",
    "MULTI_GAUSSIAN",
    "MULTI_GAUSSIAN_ELLIPSE",
    "SERSIC",
    "SERSIC_ELLIPSE",
    "SERSIC_ELLIPSE_Q_PHI",
    "CORE_SERSIC",
    "SHAPELETS",
    "SHAPELETS_POLAR",
    "SHAPELETS_POLAR_EXP",
    "SHAPELETS_ELLIPSE",
    "HERNQUIST",
    "HERNQUIST_ELLIPSE",
    "PJAFFE",
    "PJAFFE_ELLIPSE",
    "UNIFORM",
    "POWER_LAW",
    "NIE",
    "CHAMELEON",
    "DOUBLE_CHAMELEON",
    "TRIPLE_CHAMELEON",
    "INTERPOL",
    "SLIT_STARLETS",
    "SLIT_STARLETS_GEN2",
]


class LightModelBase(object):
    """Class to handle source and lens light models."""

    def __init__(self, light_model_list, smoothing=0.001, sersic_major_axis=None):
        """

        :param light_model_list: list of light models
        :param smoothing: smoothing factor for certain models (deprecated)
        :param sersic_major_axis: boolean or None, if True, uses the semi-major axis as the definition of the Sersic
         half-light radius, if False, uses the product average of semi-major and semi-minor axis. If None, uses the
         convention in the lenstronomy yaml setting (which by default is =False)
        """
        self.profile_type_list = light_model_list
        self.func_list = []
        if sersic_major_axis is None:
            sersic_major_axis = sersic_major_axis_conf
        for profile_type in light_model_list:
            if profile_type == "GAUSSIAN":
                from lenstronomy.LightModel.Profiles.gaussian import Gaussian

                self.func_list.append(Gaussian())
            elif profile_type == "GAUSSIAN_ELLIPSE":
                from lenstronomy.LightModel.Profiles.gaussian import GaussianEllipse

                self.func_list.append(GaussianEllipse())
            elif profile_type == "ELLIPSOID":
                from lenstronomy.LightModel.Profiles.ellipsoid import Ellipsoid

                self.func_list.append(Ellipsoid())
            elif profile_type == "MULTI_GAUSSIAN":
                from lenstronomy.LightModel.Profiles.gaussian import MultiGaussian

                self.func_list.append(MultiGaussian())
            elif profile_type == "MULTI_GAUSSIAN_ELLIPSE":
                from lenstronomy.LightModel.Profiles.gaussian import (
                    MultiGaussianEllipse,
                )

                self.func_list.append(MultiGaussianEllipse())
            elif profile_type == "SERSIC":
                from lenstronomy.LightModel.Profiles.sersic import Sersic

                self.func_list.append(Sersic(smoothing=smoothing))
            elif profile_type == "SERSIC_ELLIPSE":
                from lenstronomy.LightModel.Profiles.sersic import SersicElliptic

                self.func_list.append(
                    SersicElliptic(
                        smoothing=smoothing, sersic_major_axis=sersic_major_axis
                    )
                )
            elif profile_type == "SERSIC_ELLIPSE_Q_PHI":
                from lenstronomy.LightModel.Profiles.sersic import SersicElliptic_qPhi

                self.func_list.append(
                    SersicElliptic_qPhi(
                        smoothing=smoothing, sersic_major_axis=sersic_major_axis
                    )
                )
            elif profile_type == "CORE_SERSIC":
                from lenstronomy.LightModel.Profiles.sersic import CoreSersic

                self.func_list.append(
                    CoreSersic(smoothing=smoothing, sersic_major_axis=sersic_major_axis)
                )
            elif profile_type == "SHAPELETS":
                from lenstronomy.LightModel.Profiles.shapelets import ShapeletSet

                self.func_list.append(ShapeletSet())
            elif profile_type == "SHAPELETS_ELLIPSE":
                from lenstronomy.LightModel.Profiles.shapelets_ellipse import (
                    ShapeletSetEllipse,
                )

                self.func_list.append(ShapeletSetEllipse())
            elif profile_type == "SHAPELETS_POLAR":
                from lenstronomy.LightModel.Profiles.shapelets_polar import (
                    ShapeletSetPolar,
                )

                self.func_list.append(ShapeletSetPolar(exponential=False))
            elif profile_type == "SHAPELETS_POLAR_EXP":
                from lenstronomy.LightModel.Profiles.shapelets_polar import (
                    ShapeletSetPolar,
                )

                self.func_list.append(ShapeletSetPolar(exponential=True))
            elif profile_type == "HERNQUIST":
                from lenstronomy.LightModel.Profiles.hernquist import Hernquist

                self.func_list.append(Hernquist())
            elif profile_type == "HERNQUIST_ELLIPSE":
                from lenstronomy.LightModel.Profiles.hernquist import HernquistEllipse

                self.func_list.append(HernquistEllipse())
            elif profile_type == "PJAFFE":
                from lenstronomy.LightModel.Profiles.p_jaffe import PJaffe

                self.func_list.append(PJaffe())
            elif profile_type == "PJAFFE_ELLIPSE":
                from lenstronomy.LightModel.Profiles.p_jaffe import PJaffeEllipse

                self.func_list.append(PJaffeEllipse())
            elif profile_type == "UNIFORM":
                from lenstronomy.LightModel.Profiles.uniform import Uniform

                self.func_list.append(Uniform())
            elif profile_type == "POWER_LAW":
                from lenstronomy.LightModel.Profiles.power_law import PowerLaw

                self.func_list.append(PowerLaw())
            elif profile_type == "NIE":
                from lenstronomy.LightModel.Profiles.nie import NIE

                self.func_list.append(NIE())
            elif profile_type == "CHAMELEON":
                from lenstronomy.LightModel.Profiles.chameleon import Chameleon

                self.func_list.append(Chameleon())
            elif profile_type == "DOUBLE_CHAMELEON":
                from lenstronomy.LightModel.Profiles.chameleon import DoubleChameleon

                self.func_list.append(DoubleChameleon())
            elif profile_type == "TRIPLE_CHAMELEON":
                from lenstronomy.LightModel.Profiles.chameleon import TripleChameleon

                self.func_list.append(TripleChameleon())
            elif profile_type == "INTERPOL":
                from lenstronomy.LightModel.Profiles.interpolation import Interpol

                self.func_list.append(Interpol())
            elif profile_type == "SLIT_STARLETS":
                from lenstronomy.LightModel.Profiles.starlets import SLIT_Starlets

                self.func_list.append(
                    SLIT_Starlets(fast_inverse=True, second_gen=False)
                )
            elif profile_type == "SLIT_STARLETS_GEN2":
                from lenstronomy.LightModel.Profiles.starlets import SLIT_Starlets

                self.func_list.append(SLIT_Starlets(second_gen=True))
            else:
                raise ValueError(
                    "No light model of type %s found! Supported are the following models: %s"
                    % (profile_type, _MODELS_SUPPORTED)
                )
        self._num_func = len(self.func_list)

    def surface_brightness(self, x, y, kwargs_list, k=None):
        """
        :param x: coordinate in units of arcsec relative to the center of the image
        :type x: set or single 1d numpy array
        :param y: coordinate in units of arcsec relative to the center of the image
        :type y: set or single 1d numpy array
        :param kwargs_list: keyword argument list of light profile
        :param k: integer or list of integers for selecting subsets of light profiles
        """
        kwargs_list_standard = self._transform_kwargs(kwargs_list)
        x = np.array(x, dtype=float)
        y = np.array(y, dtype=float)
        flux = np.zeros_like(x)
        bool_list = self._bool_list(k=k)
        for i, func in enumerate(self.func_list):
            if bool_list[i] is True:
                out = np.array(
                    func.function(x, y, **kwargs_list_standard[i]), dtype=float
                )
                flux += out
        return flux

    def light_3d(self, r, kwargs_list, k=None):
        """Computes 3d density at radius r :param r: 3d radius units of arcsec relative
        to the center of the light profile :param kwargs_list: keyword argument list of
        light profile :param k: integer or list of integers for selecting subsets of
        light profiles."""
        kwargs_list_standard = self._transform_kwargs(kwargs_list)
        r = np.array(r, dtype=float)
        flux = np.zeros_like(r)
        bool_list = self._bool_list(k=k)
        for i, func in enumerate(self.func_list):
            if bool_list[i] is True:
                kwargs = {
                    k: v
                    for k, v in kwargs_list_standard[i].items()
                    if k not in ["center_x", "center_y"]
                }
                if self.profile_type_list[i] in [
                    "DOUBLE_CHAMELEON",
                    "CHAMELEON",
                    "HERNQUIST",
                    "HERNQUIST_ELLIPSE",
                    "PJAFFE",
                    "PJAFFE_ELLIPSE",
                    "GAUSSIAN",
                    "GAUSSIAN_ELLIPSE",
                    "MULTI_GAUSSIAN",
                    "MULTI_GAUSSIAN_ELLIPSE",
                    "NIE",
                    "POWER_LAW",
                    "TRIPLE_CHAMELEON",
                ]:
                    flux += func.light_3d(r, **kwargs)
                else:
                    raise ValueError(
                        "Light model %s does not support a 3d light distribution!"
                        % self.profile_type_list[i]
                    )
        return flux

    def total_flux(self, kwargs_list, norm=False, k=None):
        """Computes the total flux of each individual light profile. This allows to
        estimate the total flux as well as lenstronomy amp to magnitude conversions. Not
        all models are supported. The units are linked to the data to be modelled with
        associated noise properties (default is count/s).

        :param kwargs_list: list of keyword arguments corresponding to the light
            profiles. The 'amp' parameter can be missing.
        :param norm: bool, if True, computes the flux for amp=1
        :param k: int, if set, only evaluates the specific light model
        :return: list of (total) flux values attributed to each profile
        """
        kwargs_list_standard = self._transform_kwargs(kwargs_list)
        norm_flux_list = []
        bool_list = self._bool_list(k=k)
        for i, model in enumerate(self.profile_type_list):
            if bool_list[i] is True:
                if model in [
                    "SERSIC",
                    "SERSIC_ELLIPSE",
                    "INTERPOL",
                    "GAUSSIAN",
                    "GAUSSIAN_ELLIPSE",
                    "MULTI_GAUSSIAN",
                    "MULTI_GAUSSIAN_ELLIPSE",
                ]:
                    kwargs_new = kwargs_list_standard[i].copy()
                    if norm is True:
                        if model in ["MULTI_GAUSSIAN", "MULTI_GAUSSIAN_ELLIPSE"]:
                            new = {
                                "amp": np.array(kwargs_new["amp"])
                                / kwargs_new["amp"][0]
                            }
                        else:
                            new = {"amp": 1}
                        kwargs_new.update(new)
                    norm_flux = self.func_list[i].total_flux(**kwargs_new)
                    norm_flux_list.append(norm_flux)
                else:
                    raise ValueError(
                        "profile %s does not support flux normlization." % model
                    )
                #  TODO implement total flux for e.g. 'HERNQUIST', 'HERNQUIST_ELLIPSE', 'PJAFFE', 'PJAFFE_ELLIPSE',
                # 'GAUSSIAN', 'GAUSSIAN_ELLIPSE', 'POWER_LAW', 'NIE', 'CHAMELEON', 'DOUBLE_CHAMELEON' ,
                # 'TRIPLE_CHAMELEON', 'UNIFORM'
        return norm_flux_list

    def delete_interpol_caches(self):
        """Call the delete_cache method of INTERPOL profiles."""
        for i, model in enumerate(self.profile_type_list):
            if model in ["INTERPOL", "SLIT_STARLETS", "SLIT_STARLETS_GEN2"]:
                self.func_list[i].delete_cache()

    def _transform_kwargs(self, kwargs_list):
        """

        :param kwargs_list: keyword argument list as parameterised models
        :return: keyword argument list as used in the individual models
        """
        return kwargs_list

    def _bool_list(self, k=None):
        """Returns a bool list of the length of the lens models if k = None: returns
        bool list with True's if k is int, returns bool list with False's but k'th is
        True if k is a list of int, e.g. [0, 3, 5], returns a bool list with True's in
        the integers listed and False elsewhere if k is a boolean list, checks for size
        to match the numbers of models and returns it.

        :param k: None, int, or list of ints
        :return: bool list
        """
        return convert_bool_list(n=self._num_func, k=k)



__author__ = "sibirrer"

import pytest
import numpy.testing as npt
from lenstronomy.LensModel.lens_param import LensParam


class TestParam(object):
    def setup_method(self):
        self.lens_model_list = [
            "SPEP",
            "INTERPOL_SCALED",
            "SHAPELETS_CART",
            "MULTI_GAUSSIAN_KAPPA",
        ]
        self.kwargs = [
            {
                "theta_E": 1.0,
                "gamma": 2,
                "e1": 0,
                "e2": 0,
                "center_x": 0,
                "center_y": 0,
            },  # 'SPEP
            {
                "scale_factor": 1,
                "grid_interp_x": None,
                "grid_interp_y": None,
                "f_x": None,
                "f_y": None,
            },  # 'INTERPOL_SCALED'
            {
                "coeffs": [1, 1, 1, 1, 1, 1],
                "beta": 1.0,
                "center_x": 0,
                "center_y": 0,
            },  # 'SHAPELETS_CART'
            {
                "amp": [1, 2],
                "sigma": [0.5, 1],
                "center_x": 0,
                "center_y": 0,
                "scale_factor": 1,
            },  # 'MULTI_GAUSSIAN_KAPPA'
        ]
        self.kwargs_sigma = [
            {
                "theta_E_sigma": 1.0,
                "gamma_sigma": 2,
                "e1_sigma": 0.1,
                "e2_sigma": 0.1,
                "center_x_sigma": 0,
                "center_y_sigma": 0,
            },  # 'SPEP
            {"scale_factor_sigma": 1},  # 'INTERPOL_SCALED'
            {
                "coeffs_sigma": 0.1,
                "beta_sigma": 1.0,
                "center_x_sigma": 0,
                "center_y_sigma": 0,
            },  # 'SHAPELETS_CART'
            {
                "amp_sigma": [1, 1],
                "sigma_sigma": [1, 1],
                "center_x_sigma": 0,
                "center_y_sigma": 0,
                "scale_factor_sigma": 1,
            },
        ]
        self.kwargs_fixed = [
            {},
            {"grid_interp_x": None, "grid_interp_y": None, "f_x": None, "f_y": None},
            {},
            {"sigma": [1, 2]},
        ]
        self.kwargs_mean = []
        for i in range(len(self.lens_model_list)):
            kwargs_mean_k = self.kwargs[i].copy()
            kwargs_mean_k.update(self.kwargs_sigma[i])
            self.kwargs_mean.append(kwargs_mean_k)
        self.param = LensParam(
            lens_model_list=self.lens_model_list,
            kwargs_fixed=self.kwargs_fixed,
            num_images=2,
            solver_type="SHAPELETS",
            num_shapelet_lens=6,
        )
        self.param_fixed = LensParam(
            lens_model_list=self.lens_model_list,
            kwargs_fixed=self.kwargs,
            num_images=4,
            solver_type="NONE",
            num_shapelet_lens=6,
        )

    def test_get_setParams(self):
        print(self.kwargs, "kwargs")
        args = self.param.set_params(self.kwargs)
        print(args, "args")
        kwargs_new, _ = self.param.get_params(args, i=0)
        print(kwargs_new, "kwargs_new")
        args_new = self.param.set_params(kwargs_new)
        print(args_new, "args_new")
        for k in range(len(args)):
            npt.assert_almost_equal(args[k], args_new[k], decimal=8)

        args = self.param_fixed.set_params(self.kwargs)
        kwargs_new, _ = self.param_fixed.get_params(args, i=0)
        args_new = self.param_fixed.set_params(kwargs_new)
        for k in range(len(args)):
            npt.assert_almost_equal(args[k], args_new[k], decimal=8)

    def test_param_name_list(self):
        lens_model_list = [
            "SHIFT",
            "FLEXION",
            "SIS_TRUNCATED",
            "SERSIC",
            "SERSIC_ELLIPSE_POTENTIAL",
            "SERSIC_ELLIPSE_KAPPA",
            "PJAFFE",
            "PJAFFE_ELLIPSE",
            "HERNQUIST_ELLIPSE",
            "INTERPOL",
            "INTERPOL_SCALED",
            "SHAPELETS_POLAR",
            "DIPOLE",
            "GAUSSIAN_ELLIPSE_KAPPA",
            "SERSIC_ELLIPSE_KAPPA",
            "SERSIC_ELLIPSE_GAUSS_DEC",
            "NFW_ELLIPSE_GAUSS_DEC",
            "CTNFW_GAUSS_DEC",
            "GAUSSIAN_ELLIPSE_POTENTIAL",
            "MULTI_GAUSSIAN_KAPPA",
            "MULTI_GAUSSIAN_KAPPA_ELLIPSE",
        ]
        lensParam = LensParam(lens_model_list, kwargs_fixed=None)
        param_name_list = lensParam._param_name_list
        assert len(lens_model_list) == len(param_name_list)

    def test_num_params(self):
        num, list = self.param.num_param()
        assert num == 20

    def test_shapelet_solver(self):
        lens_model_list = ["SHAPELETS_CART"]
        lensParam = LensParam(
            lens_model_list,
            kwargs_fixed=[{}],
            num_images=2,
            solver_type="SHAPELETS",
            num_shapelet_lens=8,
        )
        kwargs_lens = [
            {
                "beta": 1,
                "coeffs": [0, 1, 2, 3, 4, 5, 5, 7],
                "center_x": 0,
                "center_y": 0,
            }
        ]
        args = lensParam.set_params(kwargs_lens)
        kwargs_out, i = lensParam.get_params(args, i=0)
        assert kwargs_out[0]["coeffs"][1] == 0
        assert kwargs_out[0]["beta"] == kwargs_lens[0]["beta"]
        num, param_list = lensParam.num_param()
        assert num == 8

        lensParam = LensParam(
            lens_model_list,
            kwargs_fixed=[{}],
            num_images=4,
            solver_type="SHAPELETS",
            num_shapelet_lens=8,
        )
        kwargs_lens = [
            {
                "beta": 1,
                "coeffs": [0, 1, 2, 3, 4, 5, 5, 7],
                "center_x": 0,
                "center_y": 0,
            }
        ]
        args = lensParam.set_params(kwargs_lens)
        kwargs_out, i = lensParam.get_params(args, i=0)
        assert kwargs_out[0]["coeffs"][5] == 0
        assert kwargs_out[0]["beta"] == kwargs_lens[0]["beta"]
        num, param_list = lensParam.num_param()
        assert num == 5


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import pytest
import numpy as np
import numpy.testing as npt
import unittest
from lenstronomy.LightModel.light_param import LightParam


class TestParam(object):
    def setup_method(self):
        self.light_model_list = [
            "GAUSSIAN",
            "MULTI_GAUSSIAN",
            "SERSIC",
            "SERSIC_ELLIPSE",
            "CORE_SERSIC",
            "SHAPELETS",
            "HERNQUIST",
            "HERNQUIST_ELLIPSE",
            "PJAFFE",
            "PJAFFE_ELLIPSE",
            "UNIFORM",
            "SHAPELETS",
            "SHAPELETS_POLAR_EXP",
            "SLIT_STARLETS",
        ]
        self.kwargs = [
            {"amp": 1.0, "sigma": 1, "center_x": 0, "center_y": 0},  # 'GAUSSIAN'
            {
                "amp": [1.0, 2],
                "sigma": [1, 3],
                "center_x": 0,
                "center_y": 0,
            },  # 'MULTI_GAUSSIAN'
            {
                "amp": 1,
                "R_sersic": 0.5,
                "n_sersic": 1,
                "center_x": 0,
                "center_y": 0,
            },  # 'SERSIC'
            {
                "amp": 1,
                "R_sersic": 0.5,
                "n_sersic": 1,
                "e1": 0.1,
                "e2": 0.1,
                "center_x": 0,
                "center_y": 0,
            },  # 'SERSIC_ELLIPSE'
            {
                "amp": 1,
                "R_sersic": 0.5,
                "Rb": 0.1,
                "gamma": 2.0,
                "n_sersic": 1,
                "e1": 0.1,
                "e2": 0.1,
                "center_x": 0,
                "center_y": 0,
            },
            # 'CORE_SERSIC'
            {
                "amp": [1, 1, 1],
                "beta": 0.5,
                "n_max": 1,
                "center_x": 0,
                "center_y": 0,
            },  # 'SHAPELETS'
            {"amp": 1, "Rs": 0.5, "center_x": 0, "center_y": 0},  # 'HERNQUIST'
            {
                "amp": 1,
                "Rs": 0.5,
                "center_x": 0,
                "center_y": 0,
                "e1": 0.1,
                "e2": 0.1,
            },  # 'HERNQUIST_ELLIPSE'
            {"amp": 1, "Ra": 1, "Rs": 0.5, "center_x": 0, "center_y": 0},  # 'PJAFFE'
            {
                "amp": 1,
                "Ra": 1,
                "Rs": 0.5,
                "center_x": 0,
                "center_y": 0,
                "e1": 0.1,
                "e2": 0.1,
            },  # 'PJAFFE_ELLIPSE'
            {"amp": 1},  # 'UNIFORM'
            {
                "amp": [1],
                "beta": 1,
                "n_max": 0,
                "center_x": 0,
                "center_y": 0,
            },  # 'SHAPELETS'
            {
                "amp": [1],
                "beta": 1,
                "n_max": 0,
                "center_x": 0,
                "center_y": 0,
            },  # 'SHAPELETS_POLAR_EXP'
            {
                "amp": np.ones((3 * 20**2,)),
                "n_scales": 3,
                "n_pixels": 20**2,
                "scale": 0.05,
                "center_x": 0,
                "center_y": 0,
            },  # 'SLIT_STARLETS'
        ]
        # self.kwargs_sigma = [
        #     {'amp_sigma': 1., 'sigma_sigma': 1, 'center_x_sigma': 0, 'center_y_sigma': 0},
        #     # 'GAUSSIAN'
        #     {'amp_sigma': [1., 1.], 'sigma_sigma': [1, 1], 'center_x_sigma': 0, 'center_y_sigma': 0},
        #     # 'MULTI_GAUSSIAN'
        #     {'amp_sigma': 1, 'R_sersic_sigma': 0.5, 'n_sersic_sigma': 1, 'center_x_sigma': 1, 'center_y_sigma': 1},  # 'SERSIC'
        #     {'amp_sigma': 1, 'R_sersic_sigma': 0.5, 'n_sersic_sigma': 1, 'center_x_sigma': 1,
        #      'center_y_sigma': 1, 'e1_sigma': 0.1, 'e2_sigma': 0.1},  # 'SERSIC_ELLIPSE'
        #     {'amp_sigma': 1, 'R_sersic_sigma': 0.5, 'n_sersic_sigma': 1, 'center_x_sigma': 1,
        #      'center_y_sigma': 1, 'e1_sigma': 0.1, 'e2_sigma': 0.1, 'Re_sigma': 0.01, 'gamma_sigma': 0.1},  # 'CORE_SERSIC'
        #     {'amp_sigma': [1, 1, 1], 'beta_sigma': 0.1, 'n_max_sigma': 1, 'center_x_sigma': 0, 'center_y_sigma': 0},  # 'SHAPELETS'
        #     {'amp_sigma': 1, 'Rs_sigma': 0.5, 'center_x_sigma': 0, 'center_y_sigma': 0},  # 'HERNQUIST'
        #     {'amp_sigma': 1, 'Rs_sigma': 0.5, 'center_x_sigma': 0, 'center_y_sigma': 0, 'e1_sigma': 0.1, 'e2_sigma': 0.1},  # 'HERNQUIST_ELLIPSE'
        #     {'amp_sigma': 1, 'Ra_sigma': 0.1, 'Rs_sigma': 0.5, 'center_x_sigma': 0, 'center_y_sigma': 0},  # 'PJAFFE'
        #     {'amp_sigma': 1, 'Ra_sigma': 0.1, 'Rs_sigma': 0.5, 'center_x_sigma': 0, 'center_y_sigma': 0, 'e1_sigma': 0.1, 'e2_sigma': 0.1},  # 'PJAFFE'
        #     {'amp_sigma': 0.1},  # 'UNIFORM'
        # ]
        self.kwargs_fixed = [
            {},
            {"sigma": [1, 3]},
            {},
            {},
            {},
            {"n_max": 1},
            {},
            {},
            {},
            {},
            {},
            {"n_max": 0},
            {"n_max": 0},
            {
                "n_scales": 3,
                "n_pixels": 20**2,
                "scale": 0.05,
                "center_x": 0,
                "center_y": 0,
            },
        ]
        self.kwargs_fixed_linear = [
            {},
            {"sigma": [1, 3]},
            {},
            {},
            {},
            {"n_max": 1},
            {},
            {},
            {},
            {},
            {},
            {},
            {},
            {
                "n_scales": 3,
                "n_pixels": 20**2,
                "scale": 0.05,
                "center_x": 0,
                "center_y": 0,
            },
        ]
        # self.kwargs_mean = []
        # for i in range(len(self.light_model_list)):
        #     kwargs_mean_k = self.kwargs[i].copy()
        #     #kwargs_mean_k.update(self.kwargs_sigma[i])
        #     self.kwargs_mean.append(kwargs_mean_k)
        self.param = LightParam(
            light_model_list=self.light_model_list,
            kwargs_fixed=self.kwargs_fixed,
            param_type="source_light",
            linear_solver=False,
        )
        self.param_linear = LightParam(
            light_model_list=self.light_model_list,
            kwargs_fixed=self.kwargs_fixed_linear,
            param_type="source_light",
            linear_solver=True,
        )
        self.param_fixed = LightParam(
            light_model_list=self.light_model_list,
            kwargs_fixed=self.kwargs,
            param_type="source_light",
            linear_solver=False,
        )

    def test_get_setParams(self):
        args = self.param.set_params(self.kwargs)
        kwargs_new, _ = self.param.get_params(args, i=0)
        args_new = self.param.set_params(kwargs_new)
        for k in range(len(args)):
            npt.assert_almost_equal(args[k], args_new[k], decimal=8)

        args = self.param_fixed.set_params(self.kwargs)
        kwargs_new, _ = self.param_fixed.get_params(args, i=0)
        args_new = self.param_fixed.set_params(kwargs_new)
        for k in range(len(args)):
            npt.assert_almost_equal(args[k], args_new[k], decimal=8)

        args = self.param_linear.set_params(self.kwargs)
        kwargs_new, _ = self.param_linear.get_params(args, i=0)
        args_new = self.param_linear.set_params(kwargs_new)
        for k in range(len(args)):
            npt.assert_almost_equal(args[k], args_new[k], decimal=8)

    def test_num_params(self):
        num, list = self.param.num_param()
        assert num == (66 + 1200)

    def test_param_name_list(self):
        param_name_list = self.param.param_name_list
        assert param_name_list[0][0] == "amp"

    def test_num_param_linear(self):
        kwargs_fixed = [
            {},
            {"sigma": [1, 3]},
            {},
            {},
            {},
            {"n_max": 1},
            {},
            {},
            {},
            {},
            {},
            {"n_max": 0},
            {"n_max": 0},
            {
                "n_scales": 3,
                "n_pixels": 20**2,
                "scale": 0.05,
                "center_x": 0,
                "center_y": 0,
            },
        ]
        param = LightParam(
            light_model_list=self.light_model_list,
            kwargs_fixed=kwargs_fixed,
            param_type="source_light",
            linear_solver=True,
        )

        num = param.num_param_linear()
        assert num == (16 + 1200)


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            lighModel = LightParam(light_model_list=["WRONG"], kwargs_fixed=[{}])
        with self.assertRaises(ValueError):
            lighModel = LightParam(
                light_model_list=["MULTI_GAUSSIAN"], kwargs_fixed=[{}]
            )
            lighModel.set_params(kwargs_list=[{"amp": 1, "sigma": 1}])
        with self.assertRaises(ValueError):
            lighModel = LightParam(
                light_model_list=["SHAPELETS"], kwargs_fixed=[{}], linear_solver=False
            )
            lighModel.num_param()
        with self.assertRaises(ValueError):
            lighModel = LightParam(
                light_model_list=["SHAPELETS"], kwargs_fixed=[{}], linear_solver=False
            )
            lighModel.get_params(args=[], i=0)
        with self.assertRaises(ValueError):
            lighModel = LightParam(
                light_model_list=["MULTI_GAUSSIAN"],
                kwargs_fixed=[{}],
                linear_solver=False,
            )
            lighModel.get_params(args=[1, 1, 1, 1], i=0)
        with self.assertRaises(ValueError):
            lighModel = LightParam(
                light_model_list=["SLIT_STARLETS"],
                kwargs_fixed=[{}],
                linear_solver=False,
            )
            lighModel.get_params(args=[1], i=0)
        with self.assertRaises(ValueError):
            # no fixed params provided
            lighModel = LightParam(
                light_model_list=["SLIT_STARLETS"],
                kwargs_fixed=[{}],
                linear_solver=False,
            )
            lighModel.set_params(kwargs_list=[{"amp": np.ones((3 * 20**2))}])
        with self.assertRaises(ValueError):
            # missing fixed params
            lighModel = LightParam(
                light_model_list=["SLIT_STARLETS"],
                kwargs_fixed=[{"n_scales": 3}],
                linear_solver=False,
            )
            lighModel.set_params(kwargs_list=[{"amp": np.ones((3 * 20**2))}])
        with self.assertRaises(ValueError):
            # missing fixed params
            lighModel = LightParam(
                light_model_list=["SLIT_STARLETS"],
                kwargs_fixed=[{"n_scales": 3}],
                linear_solver=False,
            )
            lighModel.num_param()
        with self.assertRaises(ValueError):
            # missing fixed params
            lighModel = LightParam(
                light_model_list=["SLIT_STARLETS"],
                kwargs_fixed=[{"amp": np.ones((3 * 20**2))}],
                linear_solver=False,
            )
            lighModel.set_params([{"n_scales": 3}])

        with self.assertRaises(ValueError):
            # missing fixed params 'n_pixels'
            lighModel = LightParam(
                light_model_list=["SLIT_STARLETS"],
                kwargs_fixed=[{}],
                linear_solver=False,
            )
            lighModel.set_params([{"n_scales": 3}])

            # missing fixed params 'n_scales'
            lighModel = LightParam(
                light_model_list=["SLIT_STARLETS"],
                kwargs_fixed=[{}],
                linear_solver=False,
            )
            lighModel.set_params([{"n_pixels": 3}])

        with self.assertRaises(ValueError):
            lighModel = LightParam(
                light_model_list=["SLIT_STARLETS"],
                kwargs_fixed=[{"amp": np.ones((3 * 20**2))}],
                linear_solver=False,
            )
            lighModel.set_params([{"n_scales": 3}])


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"


from lenstronomy.LensModel.Profiles.hernquist_ellipse_cse import HernquistEllipseCSE
from lenstronomy.LensModel.Profiles.hernquist import Hernquist

import numpy as np
import numpy.testing as npt
import pytest


class TestHernquistEllipseCSE(object):
    """Tests the Gaussian methods."""

    def setup_method(self):
        self.hernquist = Hernquist()
        self.hernquist_cse = HernquistEllipseCSE()

    def test_function(self):
        x = np.linspace(0.01, 2, 10)
        y = np.zeros_like(x)
        kwargs = {"sigma0": 2, "Rs": 2, "center_x": 0, "center_y": 0}

        f_nfw = self.hernquist.function(x, y, **kwargs)
        f_cse = self.hernquist_cse.function(x, y, e1=0, e2=0, **kwargs)
        npt.assert_almost_equal(f_cse / f_nfw, 1, decimal=5)

    def test_derivatives(self):
        x = np.linspace(0.01, 2, 10)
        y = np.zeros_like(x)
        kwargs = {"sigma0": 0.5, "Rs": 2, "center_x": 0, "center_y": 0}

        f_x_nfw, f_y_nfw = self.hernquist.derivatives(x, y, **kwargs)
        f_x_cse, f_y_cse = self.hernquist_cse.derivatives(x, y, e1=0, e2=0, **kwargs)
        npt.assert_almost_equal(f_x_cse, f_x_nfw, decimal=5)
        npt.assert_almost_equal(f_y_cse, f_y_nfw, decimal=5)

    def test_hessian(self):
        x = np.linspace(0.01, 5, 30)
        y = np.zeros_like(x)
        kwargs = {"sigma0": 0.5, "Rs": 2, "center_x": 0, "center_y": 0}

        f_xx_nfw, f_xy_nfw, f_yx_nfw, f_yy_nfw = self.hernquist.hessian(x, y, **kwargs)
        f_xx_cse, f_xy_cse, f_yx_cse, f_yy_cse = self.hernquist_cse.hessian(
            x, y, e1=0, e2=0, **kwargs
        )
        npt.assert_almost_equal(f_xx_cse / f_xx_nfw, 1, decimal=2)
        npt.assert_almost_equal(f_xy_cse, f_xy_nfw, decimal=5)
        npt.assert_almost_equal(f_yx_cse, f_yx_nfw, decimal=5)
        npt.assert_almost_equal(f_yy_cse, f_yy_nfw, decimal=5)

    def test_mass_3d_lens(self):
        R = 1
        Rs = 3
        alpha_Rs = 1
        m_3d_nfw = self.hernquist.mass_3d_lens(R, Rs, alpha_Rs)
        m_3d_cse = self.hernquist_cse.mass_3d_lens(R, Rs, alpha_Rs)
        npt.assert_almost_equal(m_3d_nfw, m_3d_cse, decimal=8)


if __name__ == "__main__":
    pytest.main()



__author__ = "dgilman"

from scipy.optimize import minimize
import numpy as np
from lenstronomy.Sampling.Samplers.pso import ParticleSwarmOptimizer
from lenstronomy.LensModel.QuadOptimizer.multi_plane_fast import MultiplaneFast
from lenstronomy.Sampling.Pool.pool import choose_pool

__all__ = ["Optimizer"]


class Optimizer(object):
    """Class which executes the optimization routines. Currently implemented as a
    particle swarm optimization followed by a downhill simplex routine.

    Particle swarm optimizer is modified from the CosmoHammer particle swarm routine
    with different convergence criteria implemented.
    """

    def __init__(
        self,
        x_image,
        y_image,
        lens_model_list,
        redshift_list,
        z_lens,
        z_source,
        parameter_class,
        astropy_instance=None,
        numerical_alpha_class=None,
        particle_swarm=True,
        re_optimize=False,
        re_optimize_scale=1.0,
        pso_convergence_mean=50000,
        foreground_rays=None,
        tol_source=1e-5,
        tol_simplex_func=1e-3,
        simplex_n_iterations=400,
    ):
        """

        :param x_image: x_image to fit (should be length 4)
        :param y_image: y_image to fit (should be length 4)
        :param lens_model_list: list of lens models for the system
        :param redshift_list: list of lens redshifts for the system
        :param z_lens: the main deflector redshift, the lens models being optimizer must be at this redshift
        :param z_source: the source redshift
        :param parameter_class: an instance of ParamClass (see documentation in QuadOptimizer.param_manager)
        :param astropy_instance: an instance of astropy to pass to the lens model
        :param numerical_alpha_class: a class to compute numerical deflection angles to pass to the lens model
        :param particle_swarm: bool, whether or not to use a PSO fit first
        :param re_optimize: bool, if True the initial spread of particles will be very tight
        :param re_optimize_scale: float, controls how tight the initial spread of particles is
        :param pso_convergence_mean: when to terminate the PSO fit
        :param foreground_rays: (optional) can pass in pre-computed foreground light rays from a previous fit
         so as to not waste time recomputing them
        :param tol_source: sigma in the source plane chi^2
        :param tol_simplex_func: tolerance for the downhill simplex optimization
        :param simplex_n_iterations: number of iterations per dimension for the downhill simplex optimization
        """

        self.fast_rayshooting = MultiplaneFast(
            x_image,
            y_image,
            z_lens,
            z_source,
            lens_model_list,
            redshift_list,
            astropy_instance,
            parameter_class,
            foreground_rays,
            tol_source,
            numerical_alpha_class,
        )

        self._tol_source = tol_source

        self._pso_convergence_mean = pso_convergence_mean

        self._param_class = parameter_class

        self._tol_simplex_func = tol_simplex_func

        self._simplex_n_iterations = simplex_n_iterations

        self._particle_swarm = particle_swarm

        self._re_optimize = re_optimize
        self._re_optimize_scale = re_optimize_scale

    def optimize(self, n_particles=50, n_iterations=250, verbose=False, threadCount=1):
        """

        :param n_particles: number of PSO particles, will be ignored if self._particle_swarm is False
        :param n_iterations: number of PSO iterations, will be ignored if self._particle_swarm is False
        :param verbose: whether to print stuff
        :param threadCount: integer; number of threads in multi-threading mode
        :return: keyword arguments that map (x_image, y_image) to the same source coordinate (source_x, source_y)
        """

        if self._particle_swarm:
            if threadCount > 1:
                pool = choose_pool(mpi=False, processes=threadCount)
            else:
                pool = None

            kwargs = self._fit_pso(n_particles, n_iterations, pool, verbose)

        else:
            kwargs = self._param_class.kwargs_lens

        kwargs_lens_final, source_penalty = self._fit_amoeba(kwargs, verbose)

        args_lens_final = self._param_class.kwargs_to_args(kwargs_lens_final)
        source_x_array, source_y_array = self.fast_rayshooting.ray_shooting_fast(
            args_lens_final
        )
        source_x, source_y = np.mean(source_x_array), np.mean(source_y_array)

        if verbose:
            print("optimization done.")
            print("Recovered source position: ", (source_x_array, source_y_array))

        return kwargs_lens_final, [source_x, source_y]

    def _fit_pso(self, n_particles, n_iterations, pool, verbose):
        """Executes the PSO."""

        low_bounds, high_bounds = self._param_class.bounds(
            self._re_optimize, self._re_optimize_scale
        )

        pso = ParticleSwarmOptimizer(
            self.fast_rayshooting.logL,
            low_bounds,
            high_bounds,
            n_particles,
            pool,
            args=[self._tol_source],
        )

        best, info = pso.optimize(
            n_iterations, verbose, early_stop_tolerance=self._pso_convergence_mean
        )

        if verbose:
            print("PSO done... ")
            print(
                "source plane chi^2: ",
                self.fast_rayshooting.source_plane_chi_square(best),
            )
            print("total chi^2: ", self.fast_rayshooting.chi_square(best))

        kwargs = self._param_class.args_to_kwargs(best)

        return kwargs

    def _fit_amoeba(self, kwargs, verbose):
        """Executes the downhill simplex."""

        args_init = self._param_class.kwargs_to_args(kwargs)

        options = {
            "adaptive": True,
            "fatol": self._tol_simplex_func,
            "maxiter": self._simplex_n_iterations * len(args_init),
        }

        method = "Nelder-Mead"

        if verbose:
            print("starting amoeba... ")

        opt = minimize(
            self.fast_rayshooting.chi_square,
            x0=args_init,
            method=method,
            options=options,
        )

        kwargs = self._param_class.args_to_kwargs(opt["x"])
        source_penalty = opt["fun"]

        return kwargs, source_penalty



__author__ = "sibirrer"
"""Tests for `prob_density` module."""

from lenstronomy.Util.prob_density import SkewGaussian, KDE1D
import lenstronomy.Util.prob_density as prob_density

import pytest
import numpy as np
import numpy.testing as npt
import unittest


class TestSkewGaussian(object):
    def setup_method(self):
        self.skewGassian = SkewGaussian()
        np.random.seed(seed=42)

    def test_pdf(self):
        x = 1
        y = self.skewGassian.pdf(x, e=0.0, w=1.0, a=0.0)
        assert y == 0.24197072451914337
        x = np.array([0, 1])
        y = self.skewGassian.pdf(x, e=0.0, w=1.0, a=0.0)
        assert y[0] == 0.3989422804014327
        assert y[1] == 0.24197072451914337

    def test_pdf_skew(self):
        x = 1
        y = self.skewGassian.pdf_skew(x, mu=1, sigma=1, skw=0.5)
        assert y == 0.39834240320473779

        y = self.skewGassian.pdf_skew(x, mu=1, sigma=1, skw=-0.5)
        assert y == 0.39834240320473779


class TestKDE1D(object):
    def setup_method(self):
        np.random.seed(seed=42)

    def gauss(self, x, mean, simga):
        return np.exp(-(((x - mean) / (simga)) ** 2) / 2) / np.sqrt(2 * np.pi) / simga

    def test_likelihood(self):
        x_array = np.linspace(0.5, 1.5, 3000)
        sigma = 0.1
        mean = 1.0
        sample = np.random.normal(loc=mean, scale=sigma, size=50000)
        kde = KDE1D(values=sample)

        x = -10
        likelihood = kde.likelihood(x)
        likelihood_true = self.gauss(x, mean=mean, simga=sigma)
        npt.assert_almost_equal(likelihood, likelihood_true, decimal=4)

        x = np.linspace(0.5, 1.5, 15)
        likelihood = kde.likelihood(x)
        likelihood_true = self.gauss(x, mean=mean, simga=sigma)
        npt.assert_almost_equal(likelihood, likelihood_true, decimal=1)


def test_compute_lower_upper_errors():
    sample = np.random.normal(loc=0, scale=1, size=100000)
    median, _ = prob_density.compute_lower_upper_errors(sample, num_sigma=0)
    npt.assert_almost_equal(median, 0, decimal=2)

    median, [[lower_sigma1, upper_sigma1]] = prob_density.compute_lower_upper_errors(
        sample, num_sigma=1
    )
    npt.assert_almost_equal(lower_sigma1, 1, decimal=2)
    npt.assert_almost_equal(upper_sigma1, 1, decimal=2)

    median, [
        [lower_sigma1, upper_sigma1],
        [lower_sigma2, upper_sigma2],
    ] = prob_density.compute_lower_upper_errors(sample, num_sigma=2)
    npt.assert_almost_equal(lower_sigma2, 2, decimal=2)
    npt.assert_almost_equal(upper_sigma2, 2, decimal=2)

    median, [
        [lower_sigma1, upper_sigma1],
        [lower_sigma2, upper_sigma2],
        [lower_sigma3, upper_sigma3],
    ] = prob_density.compute_lower_upper_errors(sample, num_sigma=3)
    npt.assert_almost_equal(lower_sigma2, 2, decimal=2)
    npt.assert_almost_equal(upper_sigma2, 2, decimal=2)

    npt.assert_almost_equal(lower_sigma3, 3, decimal=1)
    npt.assert_almost_equal(upper_sigma3, 3, decimal=1)


class TestRaise(unittest.TestCase):
    def test_raise(self):
        with self.assertRaises(ValueError):
            skewGassian = SkewGaussian()
            skewGassian.pdf_skew(x=1, mu=1, sigma=1, skw=-1)
        with self.assertRaises(ValueError):
            prob_density.compute_lower_upper_errors(sample=None, num_sigma=4)


if __name__ == "__main__":
    pytest.main()



from lenstronomy.PointSource.point_source_cached import PointSourceCached
from lenstronomy.PointSource.Types.unlensed import Unlensed
import numpy.testing as npt
import pytest


class TestPointSourceCached(object):
    def setup_method(self):
        self.ps_cached = PointSourceCached(Unlensed(), save_cache=True)
        self.ps = Unlensed()
        self.kwargs_ps = {"ra_image": [1], "dec_image": [0], "point_amp": [1]}
        self.kwargs_ps_dummy = {"ra_image": [-1], "dec_image": [10], "point_amp": [-1]}

    def test_image_position(self):
        x_img, y_img = self.ps.image_position(kwargs_ps=self.kwargs_ps)
        x_img_cached, y_img_cached = self.ps_cached.image_position(
            kwargs_ps=self.kwargs_ps
        )
        npt.assert_almost_equal(x_img_cached, x_img)
        npt.assert_almost_equal(y_img_cached, y_img)

        x_img_cached, y_img_cached = self.ps_cached.image_position(
            kwargs_ps=self.kwargs_ps_dummy
        )
        npt.assert_almost_equal(x_img_cached, x_img)
        npt.assert_almost_equal(y_img_cached, y_img)

        self.ps_cached.delete_lens_model_cache()
        x_img_cached, y_img_cached = self.ps_cached.image_position(
            kwargs_ps=self.kwargs_ps_dummy
        )
        assert x_img_cached[0] != x_img[0]

    def test_source_position(self):
        x_img, y_img = self.ps.source_position(kwargs_ps=self.kwargs_ps)
        x_img_cached, y_img_cached = self.ps_cached.source_position(
            kwargs_ps=self.kwargs_ps
        )
        npt.assert_almost_equal(x_img_cached, x_img)
        npt.assert_almost_equal(y_img_cached, y_img)

        x_img_cached, y_img_cached = self.ps_cached.source_position(
            kwargs_ps=self.kwargs_ps_dummy
        )
        npt.assert_almost_equal(x_img_cached, x_img)
        npt.assert_almost_equal(y_img_cached, y_img)

        self.ps_cached.delete_lens_model_cache()
        x_img_cached, y_img_cached = self.ps_cached.source_position(
            kwargs_ps=self.kwargs_ps_dummy
        )
        assert x_img_cached[0] != x_img[0]

    def test_image_amplitude(self):
        amp = self.ps.image_amplitude(kwargs_ps=self.kwargs_ps)
        amp_cached = self.ps_cached.image_amplitude(kwargs_ps=self.kwargs_ps)
        npt.assert_almost_equal(amp_cached, amp)

        # amplitudes are not cached!!!
        amp_cached = self.ps_cached.image_amplitude(kwargs_ps=self.kwargs_ps_dummy)
        assert amp_cached[0] != amp[0]

        self.ps_cached.delete_lens_model_cache()
        amp_cached = self.ps_cached.image_amplitude(kwargs_ps=self.kwargs_ps_dummy)
        assert amp_cached[0] != amp[0]

    def test_source_amplitude(self):
        amp = self.ps.source_amplitude(kwargs_ps=self.kwargs_ps)
        amp_cached = self.ps_cached.source_amplitude(kwargs_ps=self.kwargs_ps)
        npt.assert_almost_equal(amp_cached, amp)

        # amplitudes are not cached!!!
        amp_cached = self.ps_cached.source_amplitude(kwargs_ps=self.kwargs_ps_dummy)
        assert amp_cached[0] != amp[0]

        self.ps_cached.delete_lens_model_cache()
        amp_cached = self.ps_cached.source_amplitude(kwargs_ps=self.kwargs_ps_dummy)
        assert amp_cached[0] != amp[0]


if __name__ == "__main__":
    pytest.main()



__author__ = "sibirrer"

import numpy as np
import lenstronomy.Util.param_util as param_util
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["SPEMD"]


class SPEMD(LensProfileBase):
    """Class for smooth power law ellipse mass density profile (SPEMD). This class
    effectively performs the FASTELL calculations by Renan Barkana. The parameters are
    changed and represent a spherically averaged Einstein radius an a logarithmic 3D
    mass profile slope.

    The SPEMD mass profile is defined as follow:

    .. math::
        \\kappa(x, y) = \\frac{3-\\gamma}{2} \\left(\\frac{\\theta_{E}}{\\sqrt{q x^2 + y^2/q + s^2}} \\right)^{\\gamma-1}

    with :math:`\\theta_{E}` is the (circularized) Einstein radius,
    :math:`\\gamma` is the negative power-law slope of the 3D mass distributions, :math:`q` is the minor/major axis ratio,
    and :math:`x` and :math:`y` are defined in a coordinate system aligned with the major and minor axis of the lens.

    the FASTELL definitions are as follows:

    The parameters are position :math:`(x1,x2)`, overall factor
    (:math:`b`), power (gam), axis ratio (arat) which is <=1, core radius
    squared (:math:`s2`), and the output potential (:math:`\\phi`).
    The projected mass density distribution, in units of the
    critical density, is

    .. math::
        \\kappa(x1,x2)=b_{fastell} \\left[u2+s2\\right]^{-gam},

    with :math:`u2=\\left[x1^2+x2^2/(arat^2)\\right]`.

    The conversion from lenstronomy definitions of this class to FASTELL are:

    .. math::
        q_{fastell} \\equiv q_{lenstronomy}

    .. math::
        gam \\equiv (\\gamma-1)/2

    .. math::
        b_{fastell} \\equiv (3-\\gamma)/2. * \\left(\\theta_{E}^2 / q\\right)^{gam}

    .. math::
        s2_{fastell} = s_{lenstronomy}^2 * q
    """

    param_names = ["theta_E", "gamma", "e1", "e2", "s_scale", "center_x", "center_y"]
    lower_limit_default = {
        "theta_E": 0,
        "gamma": 0,
        "e1": -0.5,
        "e2": -0.5,
        "s_scale": 0,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "theta_E": 100,
        "gamma": 100,
        "e1": 0.5,
        "e2": 0.5,
        "s_scale": 100,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self, suppress_fastell=False):
        """"""
        try:
            from fastell4py import fastell4py

            self._fastell4py_bool = True
            self.fastell4py = fastell4py
        except:
            self._fastell4py_bool = False
            if suppress_fastell:
                ImportWarning(
                    "module fastell4py not installed. You can get it from here: "
                    "https://github.com/sibirrer/fastell4py "
                    "Make sure you have a fortran compiler such that the installation works properly."
                )
                Warning(
                    "SPEMD model outputs are replaced by zeros as fastell4py package is not installed!"
                )
            else:
                raise ImportError(
                    "module fastell4py not installed. You can get it from here: "
                    "https://github.com/sibirrer/fastell4py "
                    "Make sure you have a fortran compiler such that the installation works properly."
                )
        super(SPEMD, self).__init__()

    def function(self, x, y, theta_E, gamma, e1, e2, s_scale, center_x=0, center_y=0):
        """

        :param x: x-coordinate (angle)
        :param y: y-coordinate (angle)
        :param theta_E: Einstein radius (angle), pay attention to specific definition!
        :param gamma: logarithmic slope of the power-law profile. gamma=2 corresponds to isothermal
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale in the center of the profile (angle)
        :param center_x: x-position of lens center
        :param center_y: y-position of lens center
        :return: lensing potential
        """
        x1, x2, q_fastell, gam, s2, q, phi_G = self.param_transform(
            x, y, theta_E, gamma, e1, e2, s_scale, center_x, center_y
        )
        compute_bool = self._parameter_constraints(q_fastell, gam, s2, q)
        if self._fastell4py_bool and self.is_not_empty(x1, x2) and compute_bool:
            potential = self.fastell4py.ellipphi(x1, x2, q_fastell, gam, arat=q, s2=s2)
            n = len(np.atleast_1d(x))
            if n <= 1:
                if np.shape(x) == ():
                    return np.array(potential[0])
        else:
            potential = np.zeros_like(x1)
        return potential

    def derivatives(
        self, x, y, theta_E, gamma, e1, e2, s_scale, center_x=0, center_y=0
    ):
        """

        :param x: x-coordinate (angle)
        :param y: y-coordinate (angle)
        :param theta_E: Einstein radius (angle), pay attention to specific definition!
        :param gamma: logarithmic slope of the power-law profile. gamma=2 corresponds to isothermal
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale in the center of the profile
        :param center_x: x-position of lens center
        :param center_y: y-position of lens center
        :return: deflection angles alpha_x, alpha_y
        """
        x1, x2, q_fastell, gam, s2, q, phi_G = self.param_transform(
            x, y, theta_E, gamma, e1, e2, s_scale, center_x, center_y
        )
        compute_bool = self._parameter_constraints(q_fastell, gam, s2, q)
        if self._fastell4py_bool and self.is_not_empty(x1, x2) and compute_bool:
            f_x_prim, f_y_prim = self.fastell4py.fastelldefl(
                x1, x2, q_fastell, gam, arat=q, s2=s2
            )
        else:
            f_x_prim, f_y_prim = np.zeros_like(x1), np.zeros_like(x1)
        cos_phi = np.cos(phi_G)
        sin_phi = np.sin(phi_G)

        f_x = cos_phi * f_x_prim - sin_phi * f_y_prim
        f_y = sin_phi * f_x_prim + cos_phi * f_y_prim
        return f_x, f_y

    def hessian(self, x, y, theta_E, gamma, e1, e2, s_scale, center_x=0, center_y=0):
        """

        :param x: x-coordinate (angle)
        :param y: y-coordinate (angle)
        :param theta_E: Einstein radius (angle), pay attention to specific definition!
        :param gamma: logarithmic slope of the power-law profile. gamma=2 corresponds to isothermal
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale in the center of the profile
        :param center_x: x-position of lens center
        :param center_y: y-position of lens center
        :return: Hessian components f_xx, f_xy, f_yx, f_yy
        """
        x1, x2, q_fastell, gam, s2, q, phi_G = self.param_transform(
            x, y, theta_E, gamma, e1, e2, s_scale, center_x, center_y
        )
        compute_bool = self._parameter_constraints(q_fastell, gam, s2, q)
        if self._fastell4py_bool and self.is_not_empty(x1, x2) and compute_bool:
            (
                f_x_prim,
                f_y_prim,
                f_xx_prim,
                f_yy_prim,
                f_xy_prim,
            ) = self.fastell4py.fastellmag(x1, x2, q_fastell, gam, arat=q, s2=s2)
            n = len(np.atleast_1d(x))
            if n <= 1:
                if np.shape(x) == ():
                    f_xx_prim, f_yy_prim, f_xy_prim = (
                        np.array(f_xx_prim[0]),
                        np.array(f_yy_prim[0]),
                        np.array(f_xy_prim[0]),
                    )
        else:
            f_xx_prim, f_yy_prim, f_xy_prim = (
                np.zeros_like(x1),
                np.zeros_like(x1),
                np.zeros_like(x1),
            )
        kappa = (f_xx_prim + f_yy_prim) / 2
        gamma1_value = (f_xx_prim - f_yy_prim) / 2
        gamma2_value = f_xy_prim

        gamma1 = np.cos(2 * phi_G) * gamma1_value - np.sin(2 * phi_G) * gamma2_value
        gamma2 = +np.sin(2 * phi_G) * gamma1_value + np.cos(2 * phi_G) * gamma2_value

        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy

    def param_transform(
        self, x, y, theta_E, gamma, e1, e2, s_scale, center_x=0, center_y=0
    ):
        """Transforms parameters in the format of fastell4py.

        :param x: x-coordinate (angle)
        :param y: y-coordinate (angle)
        :param theta_E: Einstein radius (angle), pay attention to specific definition!
        :param gamma: logarithmic slope of the power-law profile. gamma=2 corresponds to
            isothermal
        :param e1: eccentricity component
        :param e2: eccentricity component
        :param s_scale: smoothing scale in the center of the profile
        :param center_x: x-position of lens center
        :param center_y: y-position of lens center
        :return: x-rotated, y-rotated, q_fastell, gam, s2, q, phi_G
        """
        phi_G, q = param_util.ellipticity2phi_q(e1, e2)
        x = np.array(x)
        y = np.array(y)
        x_shift = x - center_x
        y_shift = y - center_y
        q_fastell, gam, s2 = self.convert_params(theta_E, gamma, q, s_scale)
        cos_phi = np.cos(phi_G)
        sin_phi = np.sin(phi_G)

        x1 = cos_phi * x_shift + sin_phi * y_shift
        x2 = -sin_phi * x_shift + cos_phi * y_shift
        return x1, x2, q_fastell, gam, s2, q, phi_G

    @staticmethod
    def convert_params(theta_E, gamma, q, s_scale):
        """Converts parameter definitions into quantities used by the FASTELL fortran
        library.

        :param theta_E: Einstein radius
        :param gamma: 3D power-law slope of mass profile
        :param q: axis ratio minor/major
        :param s_scale: float, smoothing scale in the core
        :return: pre-factors to SPEMP profile for FASTELL
        """
        gam = (gamma - 1) / 2.0
        q_fastell = (3 - gamma) / 2.0 * (theta_E**2 / q) ** gam
        s2 = s_scale**2 * q
        return q_fastell, gam, s2

    @staticmethod
    def is_not_empty(x1, x2):
        """Check if float or not an empty array.

        :return: True if x1 and x2 are either floats/ints or an non-empty array, False
            if e.g. objects are []
        :rtype: bool
        """
        assert type(x1) == type(x2)

        if isinstance(x1, (list, tuple, np.ndarray)):
            if len(x1) != 0 and len(x2) != 0:
                return True
            else:
                return False
        else:
            return True

    @staticmethod
    def _parameter_constraints(q_fastell, gam, s2, q):
        """Sets bounds to parameters due to numerical stability.

        FASTELL has the following definitons:
        The parameters are position (x1,x2), overall factor
        (q), power (gam), axis ratio (arat) which is <=1, core radius
        squared (s2), and the output potential (phi).
        The projected mass density distribution, in units of the
        critical density, is kappa(x1,x2)=q [u2+s2]^(-gam), where
        u2=[x1^2+x2^2/(arat^2)].

        :param q_fastell: float, normalization of lens model, q_fastell = (3-gamma)/2. * (theta_E ** 2 / q) ** gam
        :param gam: float, slope parameter, gam = (gamma-1)/2.
        :param q: axis ratio
        :param s2: square of smoothing scale of the core
        :return: bool of whether or not to let the fastell provide to be evaluated or instead return zero(s)
        """
        if (
            q_fastell <= 0
            or q > 1
            or q < 0.01
            or gam > 0.999
            or gam < 0.001
            or not np.isfinite(q_fastell)
        ):
            return False
        return True



import numpy as np
import scipy
from lenstronomy.Util import util
from lenstronomy.Util import mask_util as mask_util
import lenstronomy.Util.multi_gauss_expansion as mge
from lenstronomy.Util import analysis_util
from lenstronomy.LensModel.lens_model_extensions import LensModelExtensions

__all__ = ["LensProfileAnalysis"]


class LensProfileAnalysis(object):
    """Class with analysis routines to compute derived properties of the lens model."""

    def __init__(self, lens_model):
        """

        :param lens_model: LensModel instance
        """
        self._lens_model = lens_model

    def effective_einstein_radius_grid(
        self,
        kwargs_lens,
        center_x=None,
        center_y=None,
        model_bool_list=None,
        grid_num=200,
        grid_spacing=0.05,
        get_precision=False,
        verbose=True,
    ):
        """Computes the radius with mean convergence=1 on a grid.

        :param kwargs_lens: list of lens model keyword arguments
        :param center_x: position of the center (if not set, is attempting to find it from the parameters kwargs_lens)
        :param center_y: position of the center (if not set, is attempting to find it from the parameters kwargs_lens)
        :param model_bool_list: list of booleans indicating the addition (=True) of a model component in computing the
         Einstein radius
        :param grid_num: integer, number of grid points to numerically evaluate the convergence and estimate the
         Einstein radius
        :param grid_spacing: spacing in angular units of the grid
        :param get_precision: If `True`, return the precision of estimated Einstein radius
        :param verbose: if True, indicates warning when Einstein radius can not be computed
        :type verbose: bool
        :return: estimate of the Einstein radius
        """
        center_x, center_y = analysis_util.profile_center(
            kwargs_lens, center_x, center_y
        )

        x_grid, y_grid = util.make_grid(numPix=grid_num, deltapix=grid_spacing)
        x_grid += center_x
        y_grid += center_y
        kappa = self._lens_model.kappa(x_grid, y_grid, kwargs_lens, k=model_bool_list)
        if self._lens_model.lens_model_list[0] in ["INTERPOL", "INTERPOL_SCALED"]:
            center_x = x_grid[kappa == np.max(kappa)][0]
            center_y = y_grid[kappa == np.max(kappa)][0]

        return einstein_radius_from_grid(
            kappa,
            x_grid,
            y_grid,
            grid_spacing,
            grid_num,
            center_x=center_x,
            center_y=center_y,
            get_precision=get_precision,
            verbose=verbose,
        )

    def effective_einstein_radius(
        self, kwargs_lens, r_min=1e-3, r_max=1e1, num_points=30
    ):
        """Numerical estimate of the Einstein radius with integral approximation of
        radial convergence profile.

        :param kwargs_lens: list of lens model keyword arguments
        :param r_min: minimum radius of the convergence integrand
        :param r_max: maximum radius of the convergence integrand (should be larger than
            Einstein radius)
        :param num_points: number of radial points in log spacing
        :return: estimate of the Einstein radius
        """
        r_array = np.logspace(np.log10(r_min), np.log10(r_max), num_points)

        # Define the integrand function for the 1D numerical integration: this is the surface mass density
        # kappa at a given radius r, multiplied by 2*pi*r to account for the circular geometry.
        kappa_r = self.radial_lens_profile(r_array, kwargs_lens, center_x=0, center_y=0)

        # here we make a finer grid interpolation in log-log space
        k_interp = scipy.interpolate.interp1d(np.log10(r_array), np.log10(kappa_r))
        r_array = np.logspace(np.log10(r_min), np.log10(r_max), num_points * 10)
        kappa_r = 10 ** k_interp(np.log10(r_array))

        # we perform the integral in logarithmic steps of the convergence
        kappa_r = np.array(kappa_r)
        kappa_r_ = (kappa_r[1:] + kappa_r[:-1]) / 2
        r_array_ = (r_array[1:] + r_array[:-1]) / 2
        dlog_r = (np.log10(r_array[2]) - np.log10(r_array[1])) * np.log(10)
        # add the mass within the innermost bin and assume it's constant
        kappa_innermost = kappa_r[0] * np.pi * r_array[0] ** 2

        # the first part is the logarithmic integrand, the second part the circle integrand
        kappa_slice = kappa_r_ * dlog_r * r_array_ * (2 * np.pi * r_array_)
        kappa_slice = np.append(kappa_innermost, kappa_slice)

        kappa_cdf = np.cumsum(kappa_slice)
        # calculate average convergence at radius
        kappa_average = kappa_cdf / (np.pi * r_array**2)

        # we interpolate as the inverse function and evaluate this function for average kappa = 1
        # (assumes monotonic decline in average convergence)
        inv_interp = scipy.interpolate.interp1d(
            np.log10(kappa_average), np.log10(r_array)
        )
        try:
            theta_e = 10 ** inv_interp(0)
        except:
            theta_e = np.nan
        return theta_e

    def local_lensing_effect(
        self, kwargs_lens, ra_pos=0, dec_pos=0, model_list_bool=None
    ):
        """Computes deflection, shear and convergence at (ra_pos,dec_pos) for those part
        of the lens model not included in the main deflector.

        :param kwargs_lens: lens model keyword argument list
        :param ra_pos: RA position where to compute the external effect
        :param dec_pos: DEC position where to compute the external effect
        :param model_list_bool: boolean list indicating which models effect to be added
            to the estimate
        :return: alpha_x, alpha_y, kappa, shear1, shear2
        """
        f_x, f_y = self._lens_model.alpha(
            ra_pos, dec_pos, kwargs_lens, k=model_list_bool
        )
        f_xx, f_xy, f_yx, f_yy = self._lens_model.hessian(
            ra_pos, dec_pos, kwargs_lens, k=model_list_bool
        )
        kappa = (f_xx + f_yy) / 2.0
        shear1 = 1.0 / 2 * (f_xx - f_yy)
        shear2 = f_xy
        return f_x, f_y, kappa, shear1, shear2

    def profile_slope(
        self,
        kwargs_lens,
        radius,
        center_x=None,
        center_y=None,
        model_list_bool=None,
        num_points=10,
    ):
        """Computes the logarithmic power-law slope of a profile. ATTENTION: this is not
        an observable!

        :param kwargs_lens: lens model keyword argument list
        :param radius: radius from the center where to compute the logarithmic slope
            (angular units
        :param center_x: center of profile from where to compute the slope
        :param center_y: center of profile from where to compute the slope
        :param model_list_bool: bool list, indicate which part of the model to consider
        :param num_points: number of estimates around the Einstein radius
        :return: logarithmic power-law slope
        """
        center_x, center_y = analysis_util.profile_center(
            kwargs_lens, center_x, center_y
        )
        x, y = util.points_on_circle(radius, num_points)
        dr = 0.01
        x_dr, y_dr = util.points_on_circle(radius + dr, num_points)

        alpha_E_x_i, alpha_E_y_i = self._lens_model.alpha(
            center_x + x, center_y + y, kwargs_lens, k=model_list_bool
        )
        alpha_E_r = np.sqrt(alpha_E_x_i**2 + alpha_E_y_i**2)
        alpha_E_dr_x_i, alpha_E_dr_y_i = self._lens_model.alpha(
            center_x + x_dr, center_y + y_dr, kwargs_lens, k=model_list_bool
        )
        alpha_E_dr = np.sqrt(alpha_E_dr_x_i**2 + alpha_E_dr_y_i**2)
        slope = np.mean(np.log(alpha_E_dr / alpha_E_r) / np.log((radius + dr) / radius))
        gamma = -slope + 2
        return gamma

    def mst_invariant_differential(
        self,
        kwargs_lens,
        radius,
        center_x=None,
        center_y=None,
        model_list_bool=None,
        num_points=10,
    ):
        """Average of the radial stretch differential in radial direction, divided by
        the radial stretch factor.

        .. math::
            \\xi = \\frac{\\partial \\lambda_{\\rm rad}}{\\partial r} \\frac{1}{\\lambda_{\\rm rad}}

        This quantity is invariant under the MST.
        The specific definition is provided by Birrer 2021. Equivalent (proportional) definitions are provided by e.g.
        Kochanek 2020, Sonnenfeld 2018.

        :param kwargs_lens: lens model keyword argument list
        :param radius: radius from the center where to compute the MST invariant differential
        :param center_x: center position
        :param center_y: center position
        :param model_list_bool: indicate which part of the model to consider
        :param num_points: number of estimates around the radius
        :return: xi
        """
        center_x, center_y = analysis_util.profile_center(
            kwargs_lens, center_x, center_y
        )
        x, y = util.points_on_circle(radius, num_points)
        ext = LensModelExtensions(lensModel=self._lens_model)
        (
            lambda_rad,
            lambda_tan,
            orientation_angle,
            dlambda_tan_dtan,
            dlambda_tan_drad,
            dlambda_rad_drad,
            dlambda_rad_dtan,
            dphi_tan_dtan,
            dphi_tan_drad,
            dphi_rad_drad,
            dphi_rad_dtan,
        ) = ext.radial_tangential_differentials(
            x, y, kwargs_lens, center_x=center_x, center_y=center_y
        )
        xi = np.mean(dlambda_rad_drad / lambda_rad)
        return xi

    def radial_lens_profile(
        self, r_list, kwargs_lens, center_x=None, center_y=None, model_bool_list=None
    ):
        """

        :param r_list: list of radii to compute the spherically averaged lens light profile
        :param center_x: center of the profile
        :param center_y: center of the profile
        :param kwargs_lens: lens parameter keyword argument list
        :param model_bool_list: bool list or None, indicating which profiles to sum over
        :return: flux amplitudes at r_list radii azimuthally averaged
        """
        center_x, center_y = analysis_util.profile_center(
            kwargs_lens, center_x, center_y
        )
        kappa_list = []
        for r in r_list:
            x, y = util.points_on_circle(r, num_points=20)
            f_r = self._lens_model.kappa(
                x + center_x, y + center_y, kwargs=kwargs_lens, k=model_bool_list
            )
            kappa_list.append(np.average(f_r))
        return kappa_list

    def multi_gaussian_lens(
        self, kwargs_lens, center_x=None, center_y=None, model_bool_list=None, n_comp=20
    ):
        """Multi-gaussian lens model in convergence space.

        :param kwargs_lens:
        :param n_comp:
        :return:
        """
        center_x, center_y = analysis_util.profile_center(
            kwargs_lens, center_x, center_y
        )
        theta_E = self.effective_einstein_radius_grid(kwargs_lens)
        r_array = np.logspace(-4, 2, 200) * theta_E
        kappa_s = self.radial_lens_profile(
            r_array,
            kwargs_lens,
            center_x=center_x,
            center_y=center_y,
            model_bool_list=model_bool_list,
        )
        amplitudes, sigmas, norm = mge.mge_1d(r_array, kappa_s, N=n_comp)
        return amplitudes, sigmas, center_x, center_y

    def mass_fraction_within_radius(
        self, kwargs_lens, center_x, center_y, theta_E, numPix=100
    ):
        """Computes the mean convergence of all the different lens model components
        within a spherical aperture.

        :param kwargs_lens: lens model keyword argument list
        :param center_x: center of the aperture
        :param center_y: center of the aperture
        :param theta_E: radius of aperture
        :return: list of average convergences for all the model components
        """
        x_grid, y_grid = util.make_grid(numPix=numPix, deltapix=2.0 * theta_E / numPix)
        x_grid += center_x
        y_grid += center_y
        mask = mask_util.mask_azimuthal(x_grid, y_grid, center_x, center_y, theta_E)
        kappa_list = []
        for i in range(len(kwargs_lens)):
            kappa = self._lens_model.kappa(x_grid, y_grid, kwargs_lens, k=i)
            kappa_mean = np.sum(kappa * mask) / np.sum(mask)
            kappa_list.append(kappa_mean)
        return kappa_list

    def convergence_peak(
        self,
        kwargs_lens,
        model_bool_list=None,
        grid_num=200,
        grid_spacing=0.01,
        center_x_init=0,
        center_y_init=0,
    ):
        """Computes the maximal convergence position on a grid and returns its
        coordinate.

        :param kwargs_lens: lens model keyword argument list
        :param model_bool_list: bool list (optional) to include certain models or not
        :return: center_x, center_y
        """
        x_grid, y_grid = util.make_grid(numPix=grid_num, deltapix=grid_spacing)
        x_grid += center_x_init
        y_grid += center_y_init

        kappa = self._lens_model.kappa(x_grid, y_grid, kwargs_lens, k=model_bool_list)

        center_x = x_grid[kappa == np.max(kappa)]
        center_y = y_grid[kappa == np.max(kappa)]
        return center_x, center_y


def einstein_radius_from_grid(
    kappa,
    x_grid,
    y_grid,
    grid_spacing,
    grid_num,
    center_x=0,
    center_y=0,
    get_precision=False,
    verbose=True,
):
    """Computes the radius with mean convergence=1.

    :param kappa: convergence calculated on a grid
    :param x_grid: x-value of grid points
    :param y_grid: y-value of grid points
    :param grid_spacing: spacing of grid points
    :param grid_num: number of grid points
    :param center_x: x-center of profile from where to measure circular averaged
        convergence
    :param center_y: y-center of profile from where to measure circular averaged
        convergence
    :param get_precision: if True, returns Einstein radius and expected numerical
        precision
    :param verbose: if True, indicates warning when Einstein radius can not be computed
    :type verbose: bool
    :return: einstein radius
    """

    r_array = np.linspace(start=0, stop=grid_num * grid_spacing / 2.0, num=grid_num * 2)
    inner_most_bin = True
    for r in r_array:
        mask = np.array(
            1 - mask_util.mask_center_2d(center_x, center_y, r, x_grid, y_grid)
        )
        sum_mask = np.sum(mask)
        if sum_mask > 0:
            kappa_mean = np.sum(kappa * mask) / np.sum(mask)
            if inner_most_bin:
                if kappa_mean < 1:
                    Warning(
                        "Central convergence value is subcritical <1 and hence an Einstein radius is ill defined."
                    )
                    if get_precision:
                        return np.nan, 0
                    else:
                        return np.nan
                inner_most_bin = False
            if kappa_mean < 1:
                if get_precision:
                    return r, r_array[1] - r_array[0]
                else:
                    return r
    if verbose:
        Warning(
            "Einstein radius could not be computed (or does not exist) for lens model."
        )
    if get_precision:
        return np.nan, 0
    else:
        return np.nan



import numpy as np
from lenstronomy.LensModel.Profiles.base_profile import LensProfileBase

__all__ = ["PJaffe"]


class PJaffe(LensProfileBase):
    """
    class to compute the DUAL PSEUDO ISOTHERMAL ELLIPTICAL MASS DISTRIBUTION
    based on Eliasdottir (2007) https://arxiv.org/pdf/0710.5636.pdf Appendix A

    Module name: 'PJAFFE';

    An alternative name is dPIED.

    The 3D density distribution is

    .. math::
        \\rho(r) = \\frac{\\rho_0}{(1+r^2/Ra^2)(1+r^2/Rs^2)}

    with :math:`Rs > Ra`.

    The projected density is

    .. math::
        \\Sigma(R) = \\Sigma_0 \\frac{Ra Rs}{Rs-Ra}\\left(\\frac{1}{\\sqrt{Ra^2+R^2}} - \\frac{1}{\\sqrt{Rs^2+R^2}} \\right)

    with

    .. math::
        \\Sigma_0 = \\pi \\rho_0 \\frac{Ra Rs}{Rs + Ra}

    In the lensing parameterization,

    .. math::
        \\sigma_0 = \\frac{\\Sigma_0}{\\Sigma_{\\rm crit}}

    """

    param_names = ["sigma0", "Ra", "Rs", "center_x", "center_y"]
    lower_limit_default = {
        "sigma0": 0,
        "Ra": 0,
        "Rs": 0,
        "center_x": -100,
        "center_y": -100,
    }
    upper_limit_default = {
        "sigma0": 10,
        "Ra": 100,
        "Rs": 100,
        "center_x": 100,
        "center_y": 100,
    }

    def __init__(self):
        """"""
        LensProfileBase.__init__(self)
        self._s = 0.0001

    def density(self, r, rho0, Ra, Rs):
        """Computes the density.

        :param r: radial distance from the center (in 3D)
        :param rho0: density normalization (see class documentation above)
        :param Ra: core radius
        :param Rs: transition radius from logarithmic slope -2 to -4
        :return: density at r
        """
        Ra, Rs = self._sort_ra_rs(Ra, Rs)
        rho = rho0 / ((1 + (r / Ra) ** 2) * (1 + (r / Rs) ** 2))
        return rho

    def density_2d(self, x, y, rho0, Ra, Rs, center_x=0, center_y=0):
        """Projected density.

        :param x: projected coordinate on the sky
        :param y: projected coordinate on the sky
        :param rho0: density normalization (see class documentation above)
        :param Ra: core radius
        :param Rs: transition radius from logarithmic slope -2 to -4
        :param center_x: center of profile
        :param center_y: center of profile
        :return: projected density
        """
        Ra, Rs = self._sort_ra_rs(Ra, Rs)
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        sigma0 = self.rho2sigma(rho0, Ra, Rs)
        sigma = (
            sigma0
            * Ra
            * Rs
            / (Rs - Ra)
            * (1 / np.sqrt(Ra**2 + r**2) - 1 / np.sqrt(Rs**2 + r**2))
        )
        return sigma

    def mass_3d(self, r, rho0, Ra, Rs):
        """Mass enclosed a 3d sphere or radius r.

        :param r: radial distance from the center (in 3D)
        :param rho0: density normalization (see class documentation above)
        :param Ra: core radius
        :param Rs: transition radius from logarithmic slope -2 to -4
        :return: M(<r)
        """
        m_3d = (
            4
            * np.pi
            * rho0
            * Ra**2
            * Rs**2
            / (Rs**2 - Ra**2)
            * (Rs * np.arctan(r / Rs) - Ra * np.arctan(r / Ra))
        )
        return m_3d

    def mass_3d_lens(self, r, sigma0, Ra, Rs):
        """Mass enclosed a 3d sphere or radius r given a lens parameterization with
        angular units.

        :param r: radial distance from the center (in 3D)
        :param sigma0: density normalization (see class documentation above)
        :param Ra: core radius
        :param Rs: transition radius from logarithmic slope -2 to -4
        :return: M(<r) in angular units (modulo critical mass density)
        """
        rho0 = self.sigma2rho(sigma0, Ra, Rs)
        return self.mass_3d(r, rho0, Ra, Rs)

    def mass_2d(self, r, rho0, Ra, Rs):
        """Mass enclosed projected 2d sphere of radius r.

        :param r: radial distance from the center in projection
        :param rho0: density normalization (see class documentation above)
        :param Ra: core radius
        :param Rs: transition radius from logarithmic slope -2 to -4
        :return: Sigma(<r)
        """
        Ra, Rs = self._sort_ra_rs(Ra, Rs)
        sigma0 = self.rho2sigma(rho0, Ra, Rs)
        m_2d = (
            2
            * np.pi
            * sigma0
            * Ra
            * Rs
            / (Rs - Ra)
            * (np.sqrt(Ra**2 + r**2) - Ra - np.sqrt(Rs**2 + r**2) + Rs)
        )
        return m_2d

    def mass_tot(self, rho0, Ra, Rs):
        """Total mass within the profile.

        :param rho0: density normalization (see class documentation above)
        :param Ra: core radius
        :param Rs: transition radius from logarithmic slope -2 to -4
        :return: total mass
        """
        Ra, Rs = self._sort_ra_rs(Ra, Rs)
        sigma0 = self.rho2sigma(rho0, Ra, Rs)
        m_tot = 2 * np.pi * sigma0 * Ra * Rs
        return m_tot

    def grav_pot(self, r, rho0, Ra, Rs):
        """Gravitational potential (modulo 4 pi G and rho0 in appropriate units)

        :param r: radial distance from the center (in 3D)
        :param rho0: density normalization (see class documentation above)
        :param Ra: core radius
        :param Rs: transition radius from logarithmic slope -2 to -4
        :return: gravitational potential (modulo 4 pi G and rho0 in appropriate units)
        """
        Ra, Rs = self._sort_ra_rs(Ra, Rs)
        pot = (
            4
            * np.pi
            * rho0
            * Ra**2
            * Rs**2
            / (Rs**2 - Ra**2)
            * (
                Rs / r * np.arctan(r / Rs)
                - Ra / r * np.arctan(r / Ra)
                + 1.0 / 2 * np.log((Rs**2 + r**2) / (Ra**2 + r**2))
            )
        )
        return pot

    def function(self, x, y, sigma0, Ra, Rs, center_x=0, center_y=0):
        """Lensing potential.

        :param x: projected coordinate on the sky
        :param y: projected coordinate on the sky
        :param sigma0: sigma0/sigma_crit (see class documentation above)
        :param Ra: core radius (see class documentation above)
        :param Rs: transition radius from logarithmic slope -2 to -4 (see class
            documentation above)
        :param center_x: center of profile
        :param center_y: center of profile
        :return: lensing potential
        """
        Ra, Rs = self._sort_ra_rs(Ra, Rs)
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        f_ = (
            -2
            * sigma0
            * Ra
            * Rs
            / (Rs - Ra)
            * (
                np.sqrt(Rs**2 + r**2)
                - np.sqrt(Ra**2 + r**2)
                + Ra * np.log(Ra + np.sqrt(Ra**2 + r**2))
                - Rs * np.log(Rs + np.sqrt(Rs**2 + r**2))
            )
        )
        return f_

    def derivatives(self, x, y, sigma0, Ra, Rs, center_x=0, center_y=0):
        """Deflection angles.

        :param x: projected coordinate on the sky
        :param y: projected coordinate on the sky
        :param sigma0: sigma0/sigma_crit (see class documentation above)
        :param Ra: core radius (see class documentation above)
        :param Rs: transition radius from logarithmic slope -2 to -4 (see class
            documentation above)
        :param center_x: center of profile
        :param center_y: center of profile
        :return: f_x, f_y
        """
        Ra, Rs = self._sort_ra_rs(Ra, Rs)
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        if isinstance(r, int) or isinstance(r, float):
            r = max(self._s, r)
        else:
            r[r < self._s] = self._s
        alpha_r = 2 * sigma0 * Ra * Rs / (Rs - Ra) * self._f_A20(r / Ra, r / Rs)
        f_x = alpha_r * x_ / r
        f_y = alpha_r * y_ / r
        return f_x, f_y

    def hessian(self, x, y, sigma0, Ra, Rs, center_x=0, center_y=0):
        """Hessian of lensing potential.

        :param x: projected coordinate on the sky
        :param y: projected coordinate on the sky
        :param sigma0: sigma0/sigma_crit (see class documentation above)
        :param Ra: core radius (see class documentation above)
        :param Rs: transition radius from logarithmic slope -2 to -4 (see class
            documentation above)
        :param center_x: center of profile
        :param center_y: center of profile
        :return: f_xx, f_xy, f_yx, f_yy
        """
        Ra, Rs = self._sort_ra_rs(Ra, Rs)
        x_ = x - center_x
        y_ = y - center_y
        r = np.sqrt(x_**2 + y_**2)
        if isinstance(r, int) or isinstance(r, float):
            r = max(self._s, r)
        else:
            r[r < self._s] = self._s
        gamma = (
            sigma0
            * Ra
            * Rs
            / (Rs - Ra)
            * (
                2
                * (
                    1.0 / (Ra + np.sqrt(Ra**2 + r**2))
                    - 1.0 / (Rs + np.sqrt(Rs**2 + r**2))
                )
                - (1 / np.sqrt(Ra**2 + r**2) - 1 / np.sqrt(Rs**2 + r**2))
            )
        )
        kappa = (
            sigma0
            * Ra
            * Rs
            / (Rs - Ra)
            * (1 / np.sqrt(Ra**2 + r**2) - 1 / np.sqrt(Rs**2 + r**2))
        )
        sin_2phi = -2 * x_ * y_ / r**2
        cos_2phi = (y_**2 - x_**2) / r**2
        gamma1 = cos_2phi * gamma
        gamma2 = sin_2phi * gamma

        f_xx = kappa + gamma1
        f_yy = kappa - gamma1
        f_xy = gamma2
        return f_xx, f_xy, f_xy, f_yy

    def _f_A20(self, r_a, r_s):
        """Equation A20 in Eliasdottir (2007)

        :param r_a: r/Ra
        :param r_s: r/Rs
        :return: f(R/a, R/s)
        """
        return r_a / (1 + np.sqrt(1 + r_a**2)) - r_s / (1 + np.sqrt(1 + r_s**2))

    def rho2sigma(self, rho0, Ra, Rs):
        """Converts 3d density into 2d projected density parameter, Equation A4 in
        Eliasdottir (2007)

        :param rho0: density normalization
        :param Ra: core radius (see class documentation above)
        :param Rs: transition radius from logarithmic slope -2 to -4 (see class
            documentation above)
        :return: projected density normalization
        """
        return np.pi * rho0 * Ra * Rs / (Rs + Ra)

    def sigma2rho(self, sigma0, Ra, Rs):
        """Inverse of rho2sigma()

        :param sigma0: projected density normalization
        :param Ra: core radius (see class documentation above)
        :param Rs: transition radius from logarithmic slope -2 to -4 (see class
            documentation above)
        :return: 3D density normalization
        """
        return (Rs + Ra) / Ra / Rs / np.pi * sigma0

    @staticmethod
    def _sort_ra_rs(Ra, Rs):
        """Sorts Ra and Rs to make sure Rs > Ra.

        :param Ra:
        :param Rs:
        :return: Ra, Rs in conventions used
        """
        if Ra >= Rs:
            Ra, Rs = Rs, Ra
        if Ra < 0.0001:
            Ra = 0.0001
        if Rs < Ra + 0.0001:
            Rs += 0.0001
        return Ra, Rs
